subsection{*Feedback Operator on Predicate Transformers*}

theory TransitionFeedback
  imports "../RefinementReactive/Refinement" "Complex"

begin

  definition grd_update :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('b \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> bool" ("[-(_)\<rightarrow>(_)-]") where
    "[-p\<rightarrow>f-] = [:x \<leadsto> y . p x \<and> y = f x:]"

  lemma "[-p\<rightarrow>f-] = [.p.] o [-f-]"
    by (simp add: grd_update_def update_def fun_eq_iff demonic_def assume_def le_fun_def)

  lemma assert_grd_update: "(\<And> x . p x \<Longrightarrow> p' x) \<Longrightarrow> {.p.} o [-p'\<rightarrow>f-] = {.p.} o [-f-]"
    by (auto simp add: fun_eq_iff grd_update_def update_def demonic_def assert_def le_fun_def)

  lemma grd_update_comp: "[-p\<rightarrow>f-] o [-q\<rightarrow>g-] = [-p \<sqinter> (q o f) \<rightarrow> g o f-]"
    by (auto simp add: fun_eq_iff grd_update_def update_def demonic_def le_fun_def)

  lemma grd_update_assert_comp: "[-p\<rightarrow>f-] o {.q.} = {. x . p x \<longrightarrow> q (f x).} o [-p \<rightarrow> f-]"
    by (auto simp add: fun_eq_iff grd_update_def update_def demonic_def assert_def le_fun_def)

  lemma grd_update_update_comp: "[-p\<rightarrow>f-] o [-g-] = [-p \<rightarrow> g o f-]"
    by (auto simp add: fun_eq_iff grd_update_def update_def demonic_def le_fun_def)

  lemma update_grd_update_comp: "[-g-] o [-p\<rightarrow>f-] = [-p o g \<rightarrow> f o g-]"
    by (auto simp add: fun_eq_iff grd_update_def update_def demonic_def le_fun_def)

  lemma grd_update_update [simp]: "[-\<top>\<rightarrow>f-] = [-f-]"
    by (simp add: grd_update_def update_def)

  lemma [simp]: "(\<exists>y. (a, y) = f (u, x)) = (a = fst (f (u, x)))"
    apply safe
    apply (metis fst_conv)
    by (rule_tac x = "snd (f (u, x))" in exI, simp)

  lemma pair_eq: "((a, b) = x) = (a = fst x \<and> b = snd x)"
    by auto

  lemma comp_exists: "(r OO r') x y = (\<exists> z . r x z \<and> r' z y)"
    by auto

  lemma comp_existsa: "(r OO r')  = (\<lambda> x y . \<exists> z . r x z \<and> r' z y)"
    by (auto simp add: fun_eq_iff assert_true_skip)

  lemma drop_assumption: "p \<Longrightarrow>True"
    by simp

  lemma fun_comp_simp: "((\<lambda>(x, y). (f x, y)) \<circ> (\<lambda>(a, b). (c b, d (a, b)))) = (\<lambda> (a, b) .  (((f o c) b), d (a, b)))"
    by (simp add: fun_eq_iff)

  lemma fun_comp_simp_b: "((\<lambda>(a::'c, b::'d). (c b, d (a, b))) \<circ> (\<lambda>(x::'a, y::'d). (f x, y))) = (\<lambda> (x, y) .  (c y, d (f x, y)))"
    by (simp add: fun_eq_iff)

  lemma fun_comp_simp_c: "((\<lambda>((c, d), a). (a, c, d)) \<circ> (\<lambda>(x, y). (case x of (a, b) \<Rightarrow> (c b, d (a, b)), f y)) \<circ> (\<lambda>(a, c, b). ((a, b), c))) = (\<lambda> (u,v,w) .  (f v, c w, d (u, w)))"
    by (simp add: fun_eq_iff)

  lemma fun_comp_simp_d: "(\<lambda>x. case case x of (c, b) \<Rightarrow> ((case x of (v, w) \<Rightarrow> f v, b), c) of (x, y) \<Rightarrow> p x \<and> p' y) = (\<lambda> (u,v) .  p (f u, v) \<and> p' u)"
    by (simp add: fun_eq_iff)

  lemma fun_comp_simp_e: "(\<lambda>x. case x of (v, w) \<Rightarrow> (c w, d (case x of (v, w) \<Rightarrow> f v, w))) = (\<lambda> (u, v) .  (c v, d (f u, v)))"
    by (simp add: fun_eq_iff)

definition "select S = {. x . (\<exists> u . prec S (u, x)).} o [:x \<leadsto> u, x' . x' = x \<and> prec S (u, x) :] o S o [:v, y \<leadsto> v' . v' = v:]"
  

  lemma selectc_spec: "select ({. p .} o [:r:]) = {. x . (\<exists> u . p (u, x)).} o [:x \<leadsto> v . \<exists> u y . p (u, x) \<and> r (u,x) (v,y) :]"
    by (simp add: select_def fun_eq_iff demonic_def assert_def le_fun_def, auto)

  lemma select_sconjunctive[simp]: "sconjunctive S \<Longrightarrow> sconjunctive (select S)"
    by (simp add: select_def)

  lemma sconjunctive_fusion[simp]: "sconjunctive S \<Longrightarrow> sconjunctive S' \<Longrightarrow> sconjunctive (S \<parallel> S')"
    by (simp add: fusion_prod)

  lemma sconjunctive_Skip[simp]: "sconjunctive Skip"
    by (simp add: Skip_def sconjunctive_def)

  lemma [simp]: "prec S = \<top> \<Longrightarrow> prec (select S) = \<top>"
    by (simp add: select_def top_fun_def)

definition "selectA S = {. x . (\<exists> u . prec S (u, x)).} o [:x \<leadsto> u, x' . x' = x \<and> prec S (u, x) :] o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:v, y \<leadsto> v' . v' = v:]"

definition "selectB S = {:x \<leadsto> u, x' . x = x':} o S o [:v, y \<leadsto> v' . v' = v:]"

definition "selectC S = {:x \<leadsto> u, x' . x = x':} o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:v, y \<leadsto> v' . v' = v:]"
  

definition "feedback S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((select S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:u,y \<leadsto> y' . y' = y:]"
  
definition "feedbackA S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((selectA S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:u,y \<leadsto> y' . y' = y:]"

definition "feedbackB S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((selectB S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:u,y \<leadsto> y' . y' = y:]"

definition "feedbackC S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((selectC S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:u,y \<leadsto> y' . y' = y:]"
  
lemma selectA_spec: "selectA ({. p .} o [:r:]) = {. x . (\<exists> u . p (u, x)).} o [:x \<leadsto> u . \<exists>  y . p (u, x) \<and> r (u,x) (u,y) :]"
  apply (simp add: selectA_def fusion_spec_demonic)
  by (simp add: selectA_def fun_eq_iff demonic_def assert_def le_fun_def fusion_spec_demonic, auto)
    
thm Prod_angelic_demonic_Skip

lemma feedbackB_spec: "feedbackB ({.p.} o [:r:]) = {:x \<leadsto> u, x' . p (u,x) \<and> (\<forall> v y . r (u,x) (v, y) \<longrightarrow> p (v,x)) \<and> x = x':} o [:u,x \<leadsto> y . \<exists> v y' . r (u,x) (v, y') \<and> r (v, x) (v, y):]"
  apply (simp add: feedbackB_def selectB_def Prod_spec_Skip fusion_spec_demonic)
  apply (subgoal_tac "{:x \<leadsto> (u, x').x = x':} \<circ> ({. p .} \<circ> [: r :]) \<circ> [:(v, y)\<leadsto>v' . v' = v:] 
      = {:x \<leadsto> (u, x'). p (u, x) \<and> x = x':} \<circ> [:u,x \<leadsto> v . \<exists> y . r (u,x) (v, y):]")
   apply (simp add: Prod_angelic_demonic_Skip)
   apply (simp add: fun_eq_iff demonic_def angelic_def assert_def le_fun_def, auto)
    apply (rule_tac x = a in exI, simp_all, auto)
   by (simp add: fun_eq_iff demonic_def angelic_def assert_def le_fun_def)
    
lemma feedbackC_spec: "feedbackC ({.p.} o [:r:]) = {:x \<leadsto> u, x' . p (u,x) \<and> (\<forall> y . r (u,x) (u, y) \<longrightarrow> p (u,x)) \<and> x = x':} o [:u,x \<leadsto> y . r (u, x) (u, y):]"
  apply (simp add: feedbackC_def selectC_def Prod_spec_Skip fusion_spec_demonic)
  apply (subgoal_tac "({:x \<leadsto> (u, x').x = x':} \<circ> ({. p .} \<circ> [: r \<sqinter> (\<lambda>(u, x) (v, y). u = v) :]) \<circ> [:(v, y)\<leadsto>v'. v' = v:])  
      = {:x \<leadsto> (u, x'). p (u, x) \<and> x = x':} \<circ> [:u,x \<leadsto> v . \<exists> y . r (u,x) (u, y) \<and> v = u:]")
   apply (simp add: Prod_angelic_demonic_Skip)
   apply (simp add: fun_eq_iff demonic_def angelic_def assert_def le_fun_def, auto)
  by (simp add: fun_eq_iff demonic_def angelic_def assert_def le_fun_def)

    
  lemma feedbackB_decomp: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> 
      feedbackB ( {. u, x . p (u, x) \<and> p' x.} o [:u, x \<leadsto> v, y . r (u, x) y \<and> r' x v:])
       = {. x . p' x \<and> (\<forall>b. r' x b \<longrightarrow> p (b,x)).} o [:x \<leadsto> y . \<exists> v . r' x v \<and> r (v, x) y:]"
    apply (unfold feedbackB_spec, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec angelic_def)
    apply auto
    by metis

  lemma [simp]: "prec S = \<top> \<Longrightarrow> prec (feedback S) = \<top>"
    by (simp add: feedback_def)

  lemma feedback_simp_a: "feedback ({.p.} o [:r:]) = 
    {. \<lambda>x. (\<exists>u. p (u, x)) \<and> (\<forall>a. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> p (a, x)) .} \<circ>
    [:x \<leadsto> y  . (\<exists> v .(\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (v, y))) \<and>  r (v, x) (v, y)):]"
    apply (simp add: feedback_def selectc_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: comp_assoc [THEN sym]  comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def)
    by (simp add: inf_fun_def)
      
lemma feedbackA_simp_a: "feedbackA ({.p.} o [:r:]) = 
    {. x. \<exists>u. p (u, x) .} \<circ> [:x \<leadsto> z. \<exists>a. p (a, x) \<and> r (a, x) (a, z):]"
    apply (simp add: feedbackA_def selectA_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
  apply (simp add: comp_assoc [THEN sym]  comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def)
    by (simp add: fun_eq_iff demonic_def assert_def le_fun_def, blast)


  lemma feedback_simp_b: "feedback ({.p.} o [-q\<rightarrow>f-]) = 
    {. \<lambda>x. (\<exists>u. p (u, x)) \<and> (\<forall>u. p (u, x) \<and> q (u, x) \<longrightarrow> p (fst (f (u, x)), x)) .} \<circ>
    [:x \<leadsto> y  . (\<exists> u . p (u, x) \<and> q (u, x) \<and> q (fst (f (u, x)), x) \<and> fst (f (u, x)) = fst (f (fst (f (u, x)), x)) \<and> y = snd (f (fst (f (u, x)), x))):]"
    by (simp add: feedback_simp_a grd_update_def fun_eq_iff demonic_def assert_def le_fun_def pair_eq, auto)

  lemma feedback_simp_c: "feedback ({.p.} o [-f-]) = 
    {. x. (\<exists>u. p (u, x)) \<and> (\<forall>u. p (u, x) \<longrightarrow> p (fst (f (u, x)), x)) .} \<circ>
    [:x \<leadsto> y  . (\<exists> u . p (u, x) \<and> fst (f (u, x)) = fst (f (fst (f (u, x)), x)) \<and> y = snd (f (fst (f (u, x)), x))):]"
    by (cut_tac p = p and f = f and q = \<top> in feedback_simp_b, simp)

  lemma feedback_simp_cc: "feedback ([-f-]) = 
    [:x \<leadsto> y  . (\<exists> u . fst (f (u, x)) = fst (f (fst (f (u, x)), x)) \<and> y = snd (f (fst (f (u, x)), x))):]"
    by (cut_tac p = \<top> and f = f in feedback_simp_c, simp add: assert_true_skip_a assert_true_skip)


  lemma feedback_test: "feedback ([-(\<lambda> (u,x) . (u , u))-]) = [:\<top>:]"
    by (simp add: feedback_simp_cc fun_eq_iff demonic_def le_fun_def)

  lemma feedback_simp_d: "feedback [:r:] = [: x \<leadsto> y . \<exists> v . r (v, x) (v, y):]"
    apply (cut_tac p = \<top> and r = r in feedback_simp_a)
    apply simp
    apply (subgoal_tac "(\<lambda>x y. \<exists>v. (\<exists>u y. r (u, x) (v, y)) \<and> r (v, x) (v, y)) = (\<lambda>x y. \<exists>v. r (v, x) (v, y))")
    by (auto simp add: fun_eq_iff assert_true_skip assert_true_skip_a )

  lemma feedback_update_simp: "feedback ( {.p.} o [- \<lambda> (u, x) . (f x, g (u, x))-])
       = {. x . p (f x, x).} o [-\<lambda> x . g (f x, x)-]"
    apply (unfold feedback_simp_c, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec update_def)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis

  lemma feedback_update_simp_x: "feedback ( {. p.} o [- \<lambda> ux . (f (snd ux), g ux)-])
       = {. x . p (f x, x).} o [-\<lambda> x . g (f x, x)-]"
    apply (unfold feedback_simp_c, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec update_def)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis

  lemma feedback_update_simp_a: "feedback ( {.p.} o [- \<lambda> (u, s, x) . (f (s, x), g (u, s, x), h (u, s, x)) -])
       = {. s, x . p ((f (s, x)), s, x).} o [-\<lambda> (s, x) . (g ((f (s, x)), s, x), h ((f (s, x)), s, x))-]"
    apply (cut_tac p = p and f = f and g = "\<lambda> (u, s, x) . (g (u, s, x), h (u, s, x))" in feedback_update_simp_x)
    apply (unfold feedback_simp_c, simp)
    by (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec update_def)

  lemma feedback_update_simp_b: "feedback ( {.p.} o [- \<lambda> (u, s, x) . (f (s, x), g (u, s, x), h (u, s, x)) -])
       = {. s, x . p ((f (s, x)), s, x).} o [-\<lambda> (s, x) . (g ((f (s, x)), s, x), h ((f (s, x)), s, x))-]"
    apply (cut_tac p = p and f = f and g = "\<lambda> (u, s, x) . (g (u, s, x), h (u, s, x))" in feedback_update_simp)
    apply (simp add: feedback_update_simp)
    apply (unfold feedback_simp_c, simp)
    by (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec update_def)

  lemma feedback_update_simp_c: "feedback ( {. (u, s, x) . p u s x .} o [- \<lambda> (u, s, x) . (f s x, g u s x, h u s x) -])
       = {. s, x . p (f s x) s x.} o [-\<lambda> (s, x) . (g (f s x) s x, h (f s x) s x)-]"
    apply (unfold feedback_simp_c, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec update_def)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis

  lemma feedback_simp_bot: "feedback (\<bottom>::(('a \<times> 'b) \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c) \<Rightarrow> bool) = \<bottom>"
    proof -
      have [simp]: "((\<bottom>::(('a \<times> 'b) \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c) \<Rightarrow> bool)) =  ({.\<bottom>::'a \<times> 'c \<Rightarrow> bool.} o [:\<bottom> :: 'a \<times>'c \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool:])"
        by (simp add: feedback_simp_a fun_eq_iff assert_def)
      show "feedback (\<bottom>::(('a \<times> 'b) \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'c) \<Rightarrow> bool) = \<bottom>"
        by (simp add: feedback_simp_a fun_eq_iff assert_def)
    qed

  lemma "A = {.p.} o [-\<lambda> (a, b) . (c b, d (a, b))-] \<Longrightarrow> B = {.p'.} o [-f-] \<Longrightarrow> feedback (A o (B ** Skip)) 
          =  {. x . p (f (c x), x) \<and> p' (c x) .} \<circ> [-\<lambda>x . d (f (c x), x)-]"
    apply (simp add: prod_assert_update_skip comp_update_update comp_update_assert update_assert_comp assert_assert_comp comp_assoc [THEN sym])
    apply (simp add: fun_comp_simp)
    apply (cut_tac p = "p \<sqinter> ((\<lambda>(x, y). p' x) \<circ> (\<lambda>(a, b). (c b, d (a, b))))" and f = "f o c" and g = d in feedback_update_simp)
    by simp

  lemma AAA: "p = p' \<Longrightarrow> (\<And> x . p x \<Longrightarrow> r x = r' x) \<Longrightarrow> {.p.} o [:r:] = {.p'.} o [:r':]"
    by (simp add: fun_eq_iff demonic_def assert_def le_fun_def, auto)

  thm feedback_simp_a
    

    
  lemma "A = {.p.} o [-\<lambda> (a, b) . (c b, d (a, b))-] \<Longrightarrow> B = {.p'.} o [-f-] \<Longrightarrow> feedback ((B ** Skip) o A) 
          =  {. x . p (f (c x), x) \<and> p' (c x) .} \<circ> [-\<lambda>x . d (f (c x), x)-]"
    apply (simp add: prod_assert_update_skip comp_update_update comp_update_assert update_assert_comp assert_assert_comp comp_assoc [THEN sym])
    apply (simp add: fun_comp_simp_b)
    apply (cut_tac p = " (\<lambda>(x, y). p' x) \<sqinter> (p \<circ> (\<lambda>(x, y). (f x, y)))" and f = "c" and g = "\<lambda> (x, y) . d (f x, y)" in feedback_update_simp)
    by (simp add: update_def demonic_def fun_eq_iff assert_def le_fun_def, auto)
 

  lemma "A = {.p.} o [-\<lambda> (a, b) . (c b, d (a, b))-] \<Longrightarrow> B = {.p'.} o [-f-] \<Longrightarrow> 
        feedback (feedback ([-\<lambda> (a,c,b) . ((a,b), c)-] o (A ** B) o [-\<lambda> ((c, d), a) . (a,c,d)-] )) =  {. x . p (f (c x), x) \<and> p' (c x) .} \<circ> [-\<lambda>x . d (f (c x), x)-]"
    apply (simp add: prod_assert_update comp_update_update comp_update_assert update_assert_comp assert_assert_comp comp_assoc [THEN sym])
    apply (simp add: fun_comp_simp_c)
    apply (cut_tac p = "(\<lambda>(x, y). p x \<and> p' y) \<circ> (\<lambda>(a, c, b). ((a, b), c)) " and f = "\<lambda> (v, w) . f v" and g = "\<lambda>(u, v, w) . (c w, d (u, w))" in feedback_update_simp)
    apply (subgoal_tac "(\<lambda>(u, x). (case x of (v, w) \<Rightarrow> f v, case (u, x) of (u, v, w) \<Rightarrow> (c w, d (u, w)))) = (\<lambda>(u, v, w). (f v, c w, d (u, w)))")
    apply simp
    apply (drule drop_assumption)
    apply (drule drop_assumption)
    apply (drule drop_assumption)
    apply (drule drop_assumption)
    apply simp
    apply (simp add: fun_comp_simp_d fun_comp_simp_e)
    apply (cut_tac p = "  \<lambda>(u, v). p (f u, v) \<and> p' u " and f = "c" and g = "\<lambda> (x, y) . d (f x, y)" in feedback_update_simp)
    apply (simp add: update_def demonic_def fun_eq_iff assert_def le_fun_def)
    by (simp_all add: fun_eq_iff)
  lemma feedback_simp_aa: "feedback ({.inpt r.} o [:r:]) = 
    {. \<lambda>x. (\<exists>u. inpt r (u, x)) \<and> (\<forall>a. (\<exists>u. inpt r (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> inpt r (a, x)).} \<circ>
    [:x \<leadsto> y  . (\<exists> v .(\<exists>u.  (\<exists>y. r (u, x) (v, y))) \<and>  r (v, x) (v, y)):]"
    apply (simp add: feedback_def selectc_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: comp_assoc [THEN sym] comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def)
    apply (simp add: inf_fun_def)
    apply (simp add: fun_eq_iff demonic_def assert_def inpt_def le_fun_def, auto)
    apply (drule_tac x = xb in spec, safe, simp)
    apply metis
    by metis


  lemma feedback_in_simp_aux: "((\<exists>u. inpt r (u, x)) \<and> (\<forall>a. (\<exists>u. inpt r (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> inpt r (a, x))) 
          = ((\<exists>u. inpt r (u, x)) \<and> (\<forall>a. (\<exists>u y. r (u, x) (a, y)) \<longrightarrow> inpt r (a, x)))"
    apply (simp add: fun_eq_iff inpt_def)
    apply auto
    apply metis
    by metis

  lemma feedback_simp_aaa: "feedback ({.inpt r.} o [:r:]) = 
    {. \<lambda>x. (\<exists>u. inpt r (u, x)) \<and> (\<forall>a. (\<exists>u. inpt r (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> inpt r (a, x)).} \<circ>
    [:x \<leadsto> y  . (\<exists> v .  r (v, x) (v, y)):]"
    apply (simp add: feedback_def selectc_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: comp_assoc [THEN sym] comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def)
    apply (simp add: inf_fun_def)
    apply (simp add: fun_eq_iff demonic_def assert_def inpt_def le_fun_def, auto)
    apply (drule_tac x = xb in spec, safe, simp)
    by metis
      
  lemma feedbackB_simp_aaaa: "feedbackB ({.inpt r.} o [:r:]) = 
   {:x \<leadsto> (u, x').inpt r (u, x) \<and> (\<forall>v. (\<exists>y. r (u, x) (v, y)) \<longrightarrow> inpt r (v, x)) \<and> x = x':} \<circ> [:(u, x)\<leadsto>y.\<exists>v. (\<exists>y'. r (u, x) (v, y')) \<and> r (v, x) (v, y):]"
    by (simp add: feedbackB_spec)

  lemma feedbackB_simp_aaaaa: "p \<le> inpt r \<Longrightarrow> feedbackB ({.p.} o [:r:]) = 
   {:x \<leadsto> (u, x'). p (u, x) \<and> (\<forall>v. (\<exists>y. r (u, x) (v, y)) \<longrightarrow> p (v, x)) \<and> x = x':} \<circ> [:(u, x)\<leadsto>y.\<exists>v. (\<exists>y'. r (u, x) (v, y')) \<and> r (v, x) (v, y):]"
    by (simp add: feedbackB_spec)
      
      
  lemma feedback_simp_aaaa: "feedback ({.inpt r.} o [:r:]) = 
    {. \<lambda>x.((\<exists>u. inpt r (u, x)) \<and> (\<forall>a. (\<exists>u y. r (u, x) (a, y)) \<longrightarrow> inpt r (a, x))).} \<circ>
    [:x \<leadsto> y  . (\<exists> v .  r (v, x) (v, y)):]"
    by (simp add: feedback_simp_aaa feedback_in_simp_aux)
 
  lemma feedback_simp_aaaaa: "p \<le> inpt r \<Longrightarrow> feedback ({.p.} o [:r:]) = 
    {. \<lambda>x.((\<exists>u. p (u, x)) \<and> (\<forall>a. (\<exists>u y. p (u, x) \<and> r (u, x) (a, y)) \<longrightarrow> p (a, x))).} \<circ>
    [:x \<leadsto> y  . (\<exists> v .  p (v, x) \<and> r (v, x) (v, y)):]"
    apply (cut_tac r = "\<lambda> x y . p x \<and> r x y" in  feedback_simp_aaaa)
    by (simp add: prec_rel)


  lemma "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> feedback ([- \<lambda> (x, y, z) . ((x, y), z) -] o (({.p.} o [:r:]) ** ({.p'.} o [:r':])) o [- \<lambda> ((x, y), z) . (x, y, z) -]) = 
    (feedback ({.p.} o [:r:])) ** ({.p'.} o [:r':])"
    apply (simp add: Prod_spec update_def comp_assoc[THEN sym] demonic_demonic OO_def demonic_assert_comp)
    apply (simp add: comp_assoc demonic_demonic OO_def)
    apply (simp add: feedback_simp_aaaaa)
    apply (simp add:  Prod_spec)
    apply (subst feedback_simp_aaaaa)
    apply (simp add: inpt_def le_fun_def)
    apply (rule AAA)
    apply (simp_all add: fun_eq_iff)
    apply safe
    apply blast
    apply (drule_tac x = aa in spec, safe)
    apply simp
    apply (drule_tac x = ua in spec, simp_all, safe)
    apply blast
    apply (simp add: le_fun_def inpt_def,blast)
    apply blast
    apply blast
    apply blast
    by blast

  lemma feedback_in_simp_a: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> 
      feedback ( {. u, x . p (u, x) \<and> p' x.} o [:u, x \<leadsto> v, y . r (u, x) y \<and> r' x v:])
       = {. x . p' x \<and> (\<forall>b. r' x b \<longrightarrow> p (b,x)).} o [:x \<leadsto> y . \<exists> v . r' x v \<and> r (v, x) y:]"
    apply (unfold feedback_simp_a, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis

  lemma feedback_in_simp_b: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> 
      feedback ( {. u, x . p (u, x) \<and> p' x.} o [:u, x \<leadsto> v, y . r (u, x) y \<and> r' x v:])
       = {. x . p' x \<and> (\<forall>b. r' x b \<longrightarrow> p (b,x)).} o [:x \<leadsto> y . \<exists> v . r' x v \<and> r (v, x) y:]"
    apply (unfold feedback_simp_a, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis
  lemma "p \<le> inpt r \<Longrightarrow> p'' \<le> inpt r'' \<Longrightarrow> feedback ( (Skip ** ({.p.} o [:r:])) o ([-\<lambda>(x,y) . (y,x)-]) o (Skip ** ({.p''.} o [:r'':])) ) 
      = ({.p.} o [:r:]) o ({.p''.} o [:r'':])"
    apply (simp add: Prod_spec update_def comp_assoc[THEN sym] demonic_demonic OO_def demonic_assert_comp Prod_Skip_spec)
    apply (simp add: comp_assoc demonic_demonic OO_def)
    apply (simp add: Prod_spec update_def comp_assoc[THEN sym] demonic_demonic OO_def demonic_assert_comp Prod_Skip_spec assert_assert_comp)
    apply (simp add: comp_assoc demonic_demonic OO_def)
    apply (simp add: feedback_simp_a)
    apply (simp add: inpt_def le_fun_def)
    apply (rule AAA, safe)
    apply (simp add: fun_eq_iff le_fun_def, safe)
    by blast+

  lemma feedback_update_simp_aaa: "(\<And> u x. fst(f(u,x)) = fst(f(undefined,x))) \<Longrightarrow>
    feedback({.p.} o [-f-]) = {.x. p(fst(f(undefined, x)), x).} \<circ> [- \<lambda>x. snd(f(fst(f(undefined,x)),x))-] "
    apply (cut_tac f = "(\<lambda>x. fst(f(undefined, x)))" and g = "(\<lambda>(u,x). snd(f(u, x)))" in feedback_update_simp)
    apply (subgoal_tac "f = (\<lambda>(u, x). (fst(f(undefined,x)), snd(f(u,x))))")
    apply (simp_all add:fun_eq_iff)
    apply (drule drop_assumption)
    apply (simp)
    by (metis pair_eq)

  lemma feedback_update_simp_bbb: "(\<And> u x. fst(f(u,x)) = fst(f(undefined,x))) \<Longrightarrow>
    feedback([-f-]) = [- \<lambda>x. snd(f(fst(f(undefined,x)),x))-] "
    apply (cut_tac p = "(\<lambda>(x, s).True)" and f = f in feedback_update_simp_aaa)
    by (simp_all add: Refinement.assert_true_skip_a )

  thm feedback_def

  thm feedback_in_simp_a
   
  definition "feedbackless S = (SOME T . \<exists> p f . S = {.p.} o [-f-] \<and> T = {.x. p(fst(f(Eps (\<lambda> u . p (u, x)), x)), x).} \<circ> [- \<lambda>x. snd(f(fst(f(Eps (\<lambda> u . p (u, x)),x)),x))-]) "

  definition "fstsome p x = Eps (\<lambda> u . p (u, x))"

  definition "fbv p f x = fst(f(fstsome p x, x))"

  definition "fb_prec p f x = p(fbv p f x, x)"
  definition "fb_func p f x = snd(f(fbv p f x, x))"

  lemma fb_prec_simp: "fb_prec p f = (\<lambda> x . p(fbv p f x, x))"
    by (simp add: fun_eq_iff fb_prec_def)

  lemma fb_func_simp: "fb_func p f = (\<lambda> x . snd(f(fbv p f x, x)))"
    by (simp add: fun_eq_iff fb_func_def)

  lemma feedbackless_update_simp_aaa: "feedbackless({.p.} o [-f-]) = {.fb_prec p f.} \<circ> [- fb_func p f -] "
      apply (simp add: feedbackless_def fbv_def fstsome_def fb_prec_simp fb_func_simp)
      apply (rule some_equality)
      apply auto [1]
      apply safe
      apply (simp add: assert_update_eq, safe)
      apply (simp add: fun_eq_iff, safe)
      apply (metis someI)
      apply (metis someI)
      by (metis someI)

  lemma "(\<And> u x. fst(f(u,x)) = fst(f(undefined,x))) \<Longrightarrow> feedback({.p.} o [-f-]) = feedbackless({.p.} o [-f-])"
    apply (simp add: feedback_update_simp_aaa feedbackless_update_simp_aaa)
    apply (subgoal_tac "(\<lambda> x. p (fst (f (undefined, x)), x)) = fb_prec p f")
    apply (subgoal_tac "(\<lambda>x. snd (f (fst (f (undefined, x)), x))) = fb_func p f")
    apply simp
    apply (simp add: fun_eq_iff fb_func_def fbv_def, auto)
    apply metis
    apply (simp add: fun_eq_iff fb_prec_def fbv_def, auto)
    apply metis
    by metis

      
  lemma feedbackless_update_simp_bbb: "feedbackless([-f-]) = [- fb_func \<top> f -] "
    apply (cut_tac p = "\<lambda> (u,x) . True" and f = f in feedbackless_update_simp_aaa)
    apply (simp_all add: Refinement.assert_true_skip_a update_eq top_fun_def fb_prec_simp fb_func_simp)
    by (simp add: split_def)

  lemma feedback_update_simp_ccc: "feedback( {.\<bottom>.} o [-f-]) = \<bottom>"
    apply (subgoal_tac "{.\<bottom>.} o [-f-] = \<bottom>")
    apply (simp add: feedback_simp_bot)
    by (simp add: assert_false_fail)



  subsubsection{*Different Feedback Attempts*}

  definition "select'' S = [:x \<leadsto> u, x' . x' = x \<and> prec S (u, x) :] o S o [:v, y \<leadsto> v' . v' = v:]"

  definition "selectb S = {: x \<leadsto> u, x'. x = x' \<and> prec S (u, x):} o S o [:v, y \<leadsto> v' . v' = v:]"

  definition "selectd S = [:x \<leadsto> u, x' . x' = x \<and> prec S (u, x) :] o S o [:v, y \<leadsto> v' . v' = v:]"

  definition "selecte S = [:x \<leadsto> u, x' . x' = x \<and> grd S (u, x) :] o S o [:v, y \<leadsto> v' . v' = v:]"

  definition "feedbackf S = {. x . (\<exists> u . prec S (u, x)).} o [:x \<leadsto> (u, x'), u' . x' = x \<and> u' = u \<and> prec S (u, x) :] 
        o (S ** Skip) o [:(v, y), u \<leadsto> (v', y') . v = u \<and> v' = v \<and> y' = y:]"

  definition "feedbackg S = [:x \<leadsto> (u, x'), u' . x' = x \<and> u' = u \<and> grd S (u, x) :] o (S ** Skip) o [:(v, y), u \<leadsto> v', y' . v = u \<and> y' = y \<and> v' = v:]"
  
  lemma selectc''_spec: "select'' ({. p .} o [:r:]) = [:x \<leadsto> v . \<exists> u y . p (u, x) \<and> r (u,x) (v,y) :]"
    by (simp add: select''_def fun_eq_iff demonic_def assert_def le_fun_def, auto)

  lemma selectcb_spec: "selectb ({. p .} o [:r:]) = {: x \<leadsto> u,x'. x = x' \<and>  p (u, x):} o [:u, x \<leadsto> v . \<exists> y . p (u, x) \<and> r (u,x) (v,y) :]"
    by (simp add: selectb_def fun_eq_iff demonic_def assert_def le_fun_def inf_fun_def angelic_def)

 lemma feedbackf_spec: "feedbackf ({. p .} o [:r:]) = {. x . (\<exists> u . p (u, x)).} o [:x \<leadsto> u, y . p (u, x) \<and> r (u,x) (u,y) :]"
    by (simp add: feedbackf_def fun_eq_iff demonic_def assert_def le_fun_def Prod_spec_Skip, auto)

  lemma feedbackg_spec: "feedbackg ({. p .} o [:r:]) = {. x . (\<forall> u . p (u, x)).} o [:x \<leadsto> u, y .r (u,x) (u,y) :]"
    by (simp add: feedbackg_def fun_eq_iff demonic_def assert_def le_fun_def Prod_spec_Skip grd_def, auto)

  lemma selectd_spec: "selectd ({. p .} o [:r:]) = [:x \<leadsto> u, x' . x' = x \<and> p (u, x) :] o [:r:] o [:v, y \<leadsto> v' . v' = v:]"
    by (simp add: selectd_def fun_eq_iff demonic_def assert_def le_fun_def)

  lemma selecte_spec: "selecte ({. p .} o [:r:]) = {.x . \<forall> u . p (u, x).} o [:x \<leadsto> v . \<exists> u y . r (u, x) (v, y):]"
    by (simp add: selecte_def fun_eq_iff demonic_def assert_def le_fun_def grd_def, auto)

 
  definition "feedback' S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((select S) ** Skip) o S o [:u,y \<leadsto> y' . y' = y:]"
  definition "feedback'' S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((select'' S) ** Skip) o S o [:u,y \<leadsto> y' . y' = y:]"


  definition "feedbacka S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((select S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:])"
  definition "feedbackb S = [:x \<leadsto> x', x'' . x' = x \<and> x'' = x:] o ((selectb S) ** Skip) o (S \<parallel> [:u, x \<leadsto> v, y . u = v:]) o [:u,y \<leadsto> y' . y' = y:]"

  lemma feedback_simp_a_a: "feedback' ({.p.} o [:r:]) = 
    {. x. (\<exists>u. p (u, x)) \<and>  (\<forall>a. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> p (a, x)) .} \<circ>
    [: \<lambda>x y. \<exists>a aa. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (aa, y))) \<and> r (aa, x) (a, y) :]"
    apply (simp add: feedback'_def selectc_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: demonic_def assert_def le_fun_def Prod_spec_Skip Prod_Skip_spec Prod_demonic_skip Prod_skip_demonic Prod_demonic Prod_spec_demonic Prod_demonic_spec
     comp_assoc [THEN sym] demonic_demonic comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def Prod_spec Fail_assert comp_existsa)
    by (simp add: inf_fun_def)

  lemma feedback_simp_a_b: "feedback'' ({.p.} o [:r:]) = 
   {. \<lambda>x. \<forall>a. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> p (a, x) .} \<circ> [: \<lambda>x y. \<exists>a aa. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (aa, y))) \<and> r (aa, x) (a, y) :]"
    apply (simp add: feedback''_def selectc''_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    by (simp add: demonic_def assert_def le_fun_def Prod_spec_Skip Prod_Skip_spec Prod_demonic_skip Prod_skip_demonic Prod_demonic Prod_spec_demonic Prod_demonic_spec
     comp_assoc [THEN sym] demonic_demonic comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def Prod_spec Fail_assert comp_existsa)

  lemma feedbackb_simp_a: "feedbackb ({.p.} o [:r:]) = 
    {:x \<leadsto> u, x' . x = x' \<and> p (u, x) \<and> (\<forall>a. ((\<exists>y. r (u, x) (a, y))) \<longrightarrow> p (a, x)) :} \<circ>
    [:u, x \<leadsto> y  . (\<exists> v .( (\<exists>y. r (u, x) (v, y))) \<and>  r (v, x) (v, y)):]"
    apply (simp add: feedbackb_def selectcb_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: demonic_def assert_def le_fun_def Prod_spec_Skip Prod_Skip_spec Prod_demonic_skip Prod_skip_demonic Prod_demonic Prod_spec_demonic Prod_demonic_spec
     comp_assoc [THEN sym] demonic_demonic comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def Prod_spec Fail_assert comp_existsa)
    apply (simp add: inf_fun_def fun_eq_iff Prod_angelic_demonic_Skip assert_def demonic_def le_fun_def angelic_def, auto)
    by metis
    
  lemma feedbackb_simp_aa: "feedbackb ({.inpt r.} o [:r:]) = 
    {:x \<leadsto> u, x' . x = x' \<and> inpt r (u, x) \<and> (\<forall>a. ((\<exists>y. r (u, x) (a, y))) \<longrightarrow> inpt r (a, x)) :} \<circ>
    [:u, x \<leadsto> y  . (\<exists> v .( (\<exists>y. r (u, x) (v, y))) \<and>  r (v, x) (v, y)):]"
    apply (simp add: feedbackb_def selectcb_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: demonic_def assert_def le_fun_def Prod_spec_Skip Prod_Skip_spec Prod_demonic_skip Prod_skip_demonic Prod_demonic Prod_spec_demonic Prod_demonic_spec
     comp_assoc [THEN sym] demonic_demonic comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def Prod_spec Fail_assert comp_existsa)
    apply (simp add: inf_fun_def fun_eq_iff Prod_angelic_demonic_Skip assert_def demonic_def le_fun_def angelic_def inpt_def, auto)
    apply metis
    by metis

  lemma feedbacka_simp_a: "feedbacka ({.p.} o [:r:]) = 
    {. \<lambda>x. (\<exists>u. p (u, x)) \<and> (\<forall>a. (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (a, y))) \<longrightarrow> p (a, x)) .} \<circ>
    [: \<lambda>x (v, y) . (\<exists>u. p (u, x) \<and> (\<exists>y. r (u, x) (v, y))) \<and> r (v, x) (v, y):]"
    apply (simp add: feedbacka_def selectc_spec Prod_spec_Skip comp_demonic_demonic comp_existsa  fusion_spec_demonic)
    apply (simp add: demonic_def assert_def le_fun_def Prod_spec_Skip Prod_Skip_spec Prod_demonic_skip Prod_skip_demonic Prod_demonic Prod_spec_demonic Prod_demonic_spec
     comp_assoc [THEN sym] demonic_demonic comp_demonic_demonic assert_assert_comp comp_demonic_assert demonic_assert_comp OO_def Prod_spec Fail_assert comp_existsa)
    apply (simp add: inf_fun_def)
    by (auto simp add: le_fun_def demonic_def fun_eq_iff assert_def)

  lemma feedback_in_simp_a_a: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> 
      feedback' ( {. u, x . p (u, x) \<and> p' x.} o [:u, x \<leadsto> v, y . r (u, x) y \<and> r' x v:])
       = {. x . p' x \<and> (\<forall>b. r' x b \<longrightarrow> p (b,x)).} o [:x \<leadsto> y . \<exists> v . r' x v \<and> r (v, x) y:]"
    apply (unfold feedback_simp_a_a, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis

  lemma feedbacka_in_simp_a: "p \<le> inpt r \<Longrightarrow> p' \<le> inpt r' \<Longrightarrow> 
      feedbacka ( {. u, x . p (u, x) \<and> p' x.} o [:u, x \<leadsto> v, y . r (u, x) y \<and> r' x v:])
       = {. x . p' x \<and> (\<forall>b. r' x b \<longrightarrow> p (b,x)).} o [:x \<leadsto> v, y . r' x v \<and> r (v, x) y:]"
    apply (unfold feedbacka_simp_a, simp)
    apply (simp add: fun_eq_iff assert_def demonic_def le_fun_def inpt_def rel_spec)
    apply safe
    apply metis
    apply metis
    apply metis
    apply metis
    by metis


  lemma feedbacka_simp_b: "feedbacka [:r:] = [: x \<leadsto> v, y . r (v, x) (v, y):]"
    apply (cut_tac p = \<top> and r = r in feedbacka_simp_a)
    apply simp
    apply (subgoal_tac "(\<lambda>x (v, y). (\<exists>u y. r (u, x) (v, y)) \<and> r (v, x) (v, y)) = ( \<lambda>x (v, y). r (v, x) (v, y))")
    by (auto simp add: assert_true_skip assert_true_skip_a )

  subsubsection{*Feedback of Decomposable Components*}
definition "decomposable r r' r'' = (\<forall> u x v y . r (u, x) (v, y) = ((r' x v) \<and> r'' (u, x) y))"
  
lemma decomposabel_iff: "(\<exists> r' r'' . decomposable r r' r'') = ((\<forall> u x v y . r (u, x) (v, y) = ((\<exists> u y . r (u, x) (v, y)) \<and> (\<exists> v . r (u, x) (v, y))) ))"
  apply safe
     apply blast
    apply blast
   apply (simp add: decomposable_def)
  apply (rule_tac x = "(\<lambda> x v . (\<exists> y u' . r (u', x) (v, y)))" in exI)
  apply (rule_tac x = "(\<lambda> (u,x) y . (\<exists> v . r (u,x) (v, y)))" in exI)
    by (unfold decomposable_def, blast)

lemma decomposable_calc: "(\<exists> r' r'' . decomposable r r' r'') \<Longrightarrow> decomposable r (\<lambda> x v . (\<exists> y u' . r (u', x) (v, y))) (\<lambda> (u,x) y . (\<exists> v . r (u,x) (v, y)))"
  by (simp add: decomposable_def, blast)

    
lemma decomposable_inpt: "decomposable r r' r'' \<Longrightarrow> inpt r (u, x) = (inpt r' x \<and> inpt r'' (u, x))"
  by (simp add: decomposable_def inpt_def)
  

lemma decomposable_feedback_trs: "decomposable r r' r'' \<Longrightarrow> feedback (trs r) 
  = {. x . inpt r' x \<and> (\<forall>b. r' x b \<longrightarrow> inpt r'' (b, x)).} \<circ> [:x\<leadsto>y.\<exists>v. r' x v \<and> r'' (v, x) y:]"
proof -
  assume [simp]: "decomposable r r' r''"
  from this have A: "r = (\<lambda> (u, x) (v, y) . (r'' (u, x) y \<and> (r' x v)))"
    by (unfold decomposable_def, auto simp add: fun_eq_iff)
  have "feedback (trs r) = feedback ({. u, x . inpt r (u, x).} \<circ> [: r :])"
    by (simp add: trs_def)
  also have "... = feedback ({.(u, x).inpt r' x \<and> inpt r'' (u, x).} \<circ> [: r :])"
    by (subst decomposable_inpt [of r r' r''], simp_all)
  also have "... = feedback ({.(u, x). inpt r'' (u, x) \<and> inpt r' x.} \<circ> [: (u, x)\<leadsto>(v, y). r'' (u, x) y \<and> r' x v :])"
    apply (subst A)
    apply (subgoal_tac "(\<lambda> (u, x).inpt r' x \<and> inpt r'' (u, x)) = (\<lambda> (u, x).inpt r'' (u, x) \<and> inpt r' x)")
    by auto
  also have "... = {. x . inpt r' x \<and> (\<forall>b. r' x b \<longrightarrow> inpt r'' (b, x)).} \<circ> [:x\<leadsto>y.\<exists>v. r' x v \<and> r'' (v, x) y:]"
    by (subst feedback_in_simp_a, simp_all)
  finally show "feedback (trs r) = {. x . inpt r' x \<and> (\<forall>b. r' x b \<longrightarrow> inpt r'' (b, x)).} \<circ> [:x \<leadsto> y.\<exists>v. r' x v \<and> r'' (v, x) y:]"
    by simp
qed

lemma spec_eq: "(\<And> x . p x = p' x) \<Longrightarrow> (\<And> x y . p x \<Longrightarrow> r x y = r' x y) \<Longrightarrow> {.p.} o [:r:] = {.p'.} o [:r':]"
  by (simp add: fun_eq_iff demonic_def assert_def le_fun_def, auto)

theorem "decomposable r r' r'' \<Longrightarrow>  feedback (trs r) 
    = trs (\<lambda> x y . (\<forall>v. r' x v \<longrightarrow> inpt r'' (v, x)) \<and> (\<exists>v. r' x v \<and> r'' (v, x) y))"
  apply (simp add: decomposable_feedback_trs )
  apply (simp add: trs_def)
  apply (rule spec_eq)
   apply (auto simp add: inpt_def)
  by blast

  
lemma [simp]: "((\<exists>u. p u x) \<and> (\<exists>v. Ex (r v)) \<and> (\<forall>a. (\<exists>u. p u x) \<and> (\<exists>v. Ex (r v)) \<and> Ex (r a) \<longrightarrow> p a x)) 
  = (((\<exists>v. Ex (r v)) \<and> (\<forall> a . Ex (r a) \<longrightarrow> p a x)))"
  by auto

definition "Decomposable r = (\<exists> r' r'' . decomposable r r' r'')"
  
definition "fst_dec r = (\<lambda> x v . \<exists> u y . r (u, x) (v, y))"
definition "snd_dec r = (\<lambda> (u, x) y . \<exists> v . r (u, x) (v, y))"
  
lemma decomosable_fst_snd: "Decomposable r = (decomposable r (fst_dec r) (snd_dec r))"
  apply (simp add: Decomposable_def decomposable_def fst_dec_def snd_dec_def, safe)
     apply blast
    apply blast
   apply blast
  apply (rule_tac x = "fst_dec r" in exI)
  apply (rule_tac x = "snd_dec r" in exI)
  by (unfold fst_dec_def snd_dec_def, blast) 

definition "state_determ r = (\<forall> x y y' s s' s'' . r (s, x) (s', y) \<and> r (s, x) (s'', y') \<longrightarrow> s' = s'')"
    
lemma decomposable_and: "decomposable r r' r'' \<Longrightarrow> decomposable (\<lambda> (u, x) (v, y) . p(u, x) \<and> r (u,x) (v, y)) r' (\<lambda> (u,x) y . p (u, x) \<and> r'' (u, x) y)"
  by (simp add: decomposable_def, auto)

end
