# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *

# Structure for a Relay block
class SimulinkRelay(SimulinkStateBlock):
	def __init__(self, xml = None, size_in = None, size_out = None, parent = None):
		SimulinkStateBlock.__init__(self, xml, size_in, size_out, parent)
		
		# set block's parameters
		self.low_limit = get_property(self.xml, "OffSwitchValue")
		self.high_limit = get_property(self.xml, "OnSwitchValue")
		self.low_value = get_property(self.xml, "OffOutputValue")
		self.high_value = get_property(self.xml, "OnOutputValue")

	def set_init(self):
		self.set_port_init_val(self.low_value)

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())

		func_type = ":: (real \<times> real => bool) => real \<times> real => bool"
		func = iBasic("Relay", [self.low_limit, self.high_limit, self.low_value, self.high_value])

		if not opt_generic:
			return "("+str(func)+func_type+")"
		
		return func

	def split_block(self):
		return [SimulinkBlockRelaySingle(self, 0), SimulinkBlockRelaySingle(self, 1)]


# Structure for Relay variation with one output
class SimulinkBlockRelaySingle(SimulinkBlock):
	def __init__(self, relay, select_output):
		SimulinkBlock.__init__(self)
		self.relay = relay
		self.select_output = select_output
		self.outport = [relay.outport[select_output]]
		self.inport = relay.inport
		self.name = relay.name + ["A", "B"][select_output]
		self.parent = relay.parent
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())

		func_type = "::(real => bool) => real \<times> real => bool"
		func =  iBasic("Relay" + ["A", "B"][self.select_output], [self.relay.low_limit, self.relay.high_limit, self.relay.low_value, self.relay.high_value])

		if not opt_generic:
			return "("+str(func)+func_type+")"
		
		return func

# Structure for a Saturation block
class SimulinkSaturate(SimulinkBlock):
	def update_arities(self):
		# set by default the arity of the block's outport to 1
		self.set_outport_arity(0,1)
		return 0
	
	def isabelle(self, opt_fun, opt_generic, flag):
		low_limit = get_property(self.xml, "LowerLimit")
		up_limit = get_property(self.xml, "UpperLimit")
		
		if not flag:
			return iBasic(self.get_qualified_name())

		func_type = ":: (real => bool) => real => bool"
		func = iBasic("Saturation", [low_limit, up_limit])
			
		if not opt_generic:
			return "("+str(func)+func_type+")"

		return func


