# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *


# Structure for a UnitDelay / Delay of step 1 block
class SimulinkDelay(SimulinkStateBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkStateBlock.__init__(self, xml, size_in, size_out, parent)
		self.sample_dt = get_property(self.xml, "SampleTime")
		
	def split_block(self):
		n = len(self.inport)
		if n % 2 == 1:
			raise Exception('\nError:\n  An error has occured while building the delay block '+self.get_qualified_name()+'.\n  The block does not have even number of input ports.\n')
		if len(self.outport) != n:
			raise Exception('\nError:\n  An error has occured while building the delay block '+self.get_qualified_name()+'.\n  The block must have the same number of input and output ports.\n')
		k = n / 2
		skip_blocks = []
		for i in range(k):
			for j in [k, k + i]:
				assert(self.inport[j].get_arity() == 1 and self.inport[j].get_arity() == 1)
			skip_blocks.append(SimulinkDelaySingle(self, self.inport[i], self.outport[k + i], i))
			skip_blocks.append(SimulinkDelaySingle(self, self.inport[k + i], self.outport[i], k+i))
		return skip_blocks

	def set_init(self):
		val = get_property(self.xml, "InitialCondition")
		self.set_port_init_val(val)

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		n = len(self.inport)
		if n % 2 == 1:
			raise Exception('\nError:\n  An error has occured while building the delay block '+self.get_qualified_name()+'.\n  The block does not have even number of input ports.\n')
		n = n / 2
		outport = self.inport[n:]
		outport.extend(self.inport[:n])
		return iSwitch([self.inport], [outport], opt_fun)
		
		
# Structure for Delay variation with one output
class SimulinkDelaySingle(SimulinkBlock):
	def __init__(self, ud, inport, outport,i):
		SimulinkBlock.__init__(self)
		self.inport = [inport]
		self.outport = [outport]
		self.name = ud.name + "_" + str(i)
		self.parent = ud.parent
		self.sample_dt = ud.sample_dt
		self.type = ud.type
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
		return iSkip()

# TODO: code Simulink Memory block
