# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *


# Structure for a Lookup1D block
# !!! Instantiated for the Lookup block defined in afcs_m2.slx
class SimulinkLookup1D(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
		
		breakpoints = [0, 0.05, 0.1, 0.15000000000000002, 0.2, 0.25, 0.30000000000000004, 0.35000000000000003, 0.4, 0.45, 0.5, 0.55, 0.6, 0.6499999999999999, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95, 1]
		values = [0, 0.4, 0.8, 0.97, 1, 0.98, 0.96, 0.94, 0.92, 0.9, 0.88, 0.855, 0.83, 0.81, 0.79, 0.77, 0.75, 0.73, 0.72, 0.71, 0.7]
		n = len(values)
		
		inpt = self.inport[0].get_var_name()
		
		expr_init = "If %s < %s Then slope_val %s %s %s %s %s " %(inpt, breakpoints[1], inpt, breakpoints[0], breakpoints[1], values[0], values[1])
		
		expr_aux = []
		for i in range(2, n-2):
			expr_aux.append("Else If %s < %s Then slope_val %s %s %s %s %s " %(inpt, breakpoints[i], inpt, breakpoints[i-1], breakpoints[i], values[i-1], values[i]))
		expr_interm = " ".join(expr_aux)
		
		expr_end = "Else slope_val %s %s %s %s %s " %(inpt, breakpoints[n-2], breakpoints[n-1], values[n-2], values[n-1])
		
		expr = expr_init + expr_interm + expr_end
		
		return iFunc([inpt], expr, opt_fun)
		
