# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *


# Simulink Outport is defined by the generic SimulinkInOutPort class (simulink_ports.py file)


# Structure for a Scope block
class SimulinkScope(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		#the number of input ports is already defined in the xml file, should not be modified; 
		#only the output should be created
		self.create_ports(None, 1) 
	
	def update_arities(self):
		flag = 0
		if len(self.inport) == 1:
			if self.get_outport_arity(0) == 0:
				self.set_outport_arity(0, 1)
			if self.get_inport_arity(0) != self.get_outport_arity(0) and self.get_inport_arity(0) != 0:
				self.set_outport_arity(0, self.get_inport_arity(0))
				flag = 1
		else:
			s = 0
			for p in self.inport:
				if p.get_arity() == 0:
					p.set_arity(1)
					flag = 1
				s += p.get_arity()
			
			# update output port arity if needed and flag it
			if s != self.get_outport_arity(0):
				flag = 1
				self.set_outport_arity(0, s)
		
		return flag

	def split_block(self):
		if len(self.inport) != len(self.outport):
			raise Exception('\nError:\n  An error has occured while building the internal model.\n  The Scope '+self.get_qualified_name()+' has an uneven number of input and output ports.\n')
		res = []
		for i in range(len(self.inport)):
			inpt = self.inport[i]
			outpt = self.outport[i]
			res.append(SimulinkScopeSingle(self, inpt, outpt, i))
		return res


# Structure for Scope variant with one output
class SimulinkScopeSingle(SimulinkBlock):
	def __init__(self, scope, inport, outport, index):
		SimulinkBlock.__init__(self)
		self.name = scope.get_name() + str(index)
		self.type = "Scope"
		self.inport = [inport]
		self.outport = [outport]
		self.parent = scope.parent


# Structure for a Terminator block
# TODO: study better terminators
class SimulinkTerminator(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)

