# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *
from utilities.simulink_globals import *


# Simulink Inport is defined by the generic SimulinkInOutPort class (simulink_ports.py file)


# TODO: make a list with all the possible translations of a Constant
# based on the output type if set and the value it carries

# Structure for a Constant block        
class SimulinkConstant(SimulinkBlock):
	booleans = {"true": "True", "false": "False"}
	index = 0
	
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		self.val = get_property(self.xml, 'Value')
		self.eval = None
		self.out_type = get_property(self.xml,'OutDataTypeStr')
		if not (self.out_type in conv_types.keys()):
			self.out_type = ""
		else:
			self.out_type=conv_types[self.out_type]

		#if self.out_type=='bool':
		#	self.update_const_params(False)

		if self.val.lower() in SimulinkConstant.booleans.keys():
			self.val = SimulinkConstant.booleans[self.val.lower()]
			
	def update_ports_delete(self):
		pass
		
	def update_arities(self):
		# set by default the arity of a Constant's outport to 1
		self.set_outport_arity(0,1) 
		return 0

	def update_const_params(self, opt_const, opt_generic):
		if self.get_params_list() == []:
			if (opt_generic) or (opt_const and self.out_type=="") or ((not opt_generic) and self.out_type=='bool' and (self.val.lower() in SimulinkConstant.booleans.keys())):
				p = "vc_" + str(SimulinkConstant.index)
				self.set_params(p)
				SimulinkConstant.index +=1 
				#self.set_params("TYPE('a::MyNumber)")

	def update_const_vals(self, cvals):
		try:
			self.eval = cvals[self.val]
		except:
			pass

	def get_def_params(self, opt_fun, opt_generic):
		params = self.get_params()
		'''		
		if params:
			if self.out_type in conv_types.keys() and opt_not_generic:
				"""if self.out_type == "boolean" and (self.val.lower() in SimulinkConstant.booleans.keys()):
					return ""
				elif self.out_type == "boolean":
					return "(%s::'a::simulink)" % params"""
				if self.out_type == "boolean":
					return "(%s::'a::simulink)" % params
				else:
					return "(%s::%s)" %(params, conv_types[self.out_type])
				#return "(%s::%s)" %(params, conv_types[self.out_type])
			else:
				return "(%s::'a::simulink)" % params
		'''		
		if params:
			return "(%s::'a::simulink)" % params
		else:
			return params

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			name = self.get_qualified_name()
			params = self.get_params_list()
		else:
			name = "Constant"
			params = []
			if not self.eval:
				self.eval=""
			
			# Case 3: opt_generic is set
			if opt_generic:
				# Case 3.2: the constant value is Boolean
				if (self.val.lower() in SimulinkConstant.booleans.keys()) or (self.eval.lower() in SimulinkConstant.booleans.keys()):
					params = ["if %s then 1::'a else 0" %self.val]
				else:
					# Case 3.1: the constant output type is Boolean
					if self.out_type == 'bool':
						params = ["if ((%s)::'a) \<noteq> 0 then 1::'a else 0" %self.val]
					# Case 3.1: the constant output type is numerical or not set 
					else:
						params = ["%s::'a" %self.val]
			# Cases 1 and 2 that require parameter generation
			elif self.get_params():
				# Cases 1.1 and 2.1: the constant output type is Boolean
				if self.out_type == 'bool':
					return ["(%s::'a) \<noteq> 0" %self.val]
				elif self.out_type == "":
					# Case 2.2: the constant output type is not set
					if (self.val.lower() in SimulinkConstant.booleans.keys()) or (self.eval.lower() in SimulinkConstant.booleans.keys()):
						params = ["if %s then 1::'a else 0" %self.val]
					# Case 2.1: the constant output type is not set					
					else:
						params = ["%s::'a" %self.val]
			# Cases 1 and 2 that do not require parameter generation
			else:
				# Cases 1.1 and 1.2: the constant output type is not set
				if self.out_type == "":
					params = ["%s" %self.val]
				# Cases 1.2 and 2.2: the constant value is Boolean
				elif (self.val.lower() in SimulinkConstant.booleans.keys()) or (self.eval in SimulinkConstant.booleans.keys()):
					# Cases 1.2 and 2.2: the constant output type is Boolean
					if self.out_type == 'bool':
						params = ["%s" %self.val]
					# Case 1.2 and 2.2: the constant output type is numerical
					else:
						params = ["if %s then 1::%s else 0" %(self.val, self.out_type)]
				# Cases 1 and 2: the rest
				else:
					params = ["%s::%s" %(self.val, self.out_type)]
			'''
			if not opt_not_generic:
				if self.get_params():
					# to check the following lines
					if self.val.lower() in SimulinkConstant.booleans.keys() or self.out_type == 'boolean':
						params = ["if (%s::'a) \<noteq> 0 then 1::'a else 0" %self.val]
					else:
						params = ["%s::'a" %self.val]
				else:
					params = ["%s" % self.val]
			elif self.out_type=='boolean' and not( self.val.lower() in SimulinkConstant.booleans.keys()):
				params = ["(%s::'a) \<noteq> 0" % self.val]
			elif self.out_type=='boolean':
				params = ["%s" %self.val]
			elif self.get_params():
				if self.out_type in conv_types.keys():
					params = ["%s::%s" % (self.val, conv_types[self.out_type])]
				else:
					params = ["%s::'a" % self.val]
			else:
				if self.out_type in conv_types.keys():
					params = ["%s::%s" % (self.val, conv_types[self.out_type])]
				else:
					params = ["%s" %self.val]
			'''		
			'''
			if self.get_params():
				params = ["%s::'a" % self.val]
			else:
				params = ["%s" % self.val]
			'''
		return iBasic(name, params)


# Structure for a PulseGenerator block
class SimulinkDiscretePulseGeneratorOld(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")
		
		# set block's parameters
		self.amplitude_val = get_property(self.xml, "Amplitude")
		self.period_val = get_property(self.xml, "Period")
		self.pulsewidth_val = float(get_property(self.xml, "PulseWidth")) / 100
		self.phasedelay_val = get_property(self.xml, "PhaseDelay")
	
	def add_state_ports(self):
		pi = Port(self.get_id(), 0, "sin", 2)
		self.inport.append(pi)
		po = Port(self.get_id(), 0, "sout", 2)
		self.outport.append(po)
		
	def update_arities(self):
		self.outport[0].set_arity(1)
		return 0
	
	def split_block(self):
		return [SimulinkDiscretePulseGeneratorOldSingle(self, 0), SimulinkDiscretePulseGeneratorOldSingle(self, 1), SimulinkDiscretePulseGeneratorOldSingle(self, 2)]
	
	def set_init(self):
		for p in self.inport:
			if p.get_direction() == "sin":
				p.set_value(0.0)
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		func_type = "::(real \<times> real \<times> real => bool) => real \<times> real => bool"
		func = iBasic("PulseGeneratorT", [self.period_val, self.phasedelay_val, self.pulsewidth_val, self.amplitude_val] + self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
				
		return func

# Structure for PulseGeneratorOld projection
class SimulinkDiscretePulseGeneratorOldSingle(SimulinkBlock):
	def __init__(self, pulsegen, select_output):
		SimulinkBlock.__init__(self)

		'''
		if select_output != 1:
			self.params = pulsegen.get_params_list()
		else:
			self.params = []
		'''
		
		self.params = pulsegen.get_params_list()
		self.pulsegen = pulsegen
		self.select_output = select_output
		self.outport = [pulsegen.outport[select_output]]
		if select_output == 1:
			self.inport = [pulsegen.inport[0]]
		else:
			self.inport = pulsegen.inport
		self.parent = pulsegen.parent
		self.name = pulsegen.name + ["A", "B", "C"][select_output]
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
			
		if self.select_output == 0:
			func_type="::(real => bool) => real \<times> real => bool"
			func = iBasic("PulseGeneratorTA", [self.pulsegen.period_val, self.pulsegen.phasedelay_val, self.pulsegen.pulsewidth_val, self.pulsegen.amplitude_val] + self.get_params_list())
		elif self.select_output == 2:
			func_type="::(real => bool) => real \<times> real => bool"
			func = iBasic("PulseGeneratorTC", [self.pulsegen.period_val, self.pulsegen.phasedelay_val, self.pulsegen.pulsewidth_val] + self.get_params_list())
		else:
			func_type="::(real => bool) => real => bool"
			func = iBasic("PulseGeneratorTB", self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
						
		return func

# Structure for a PulseGenerator block
class SimulinkDiscretePulseGenerator(SimulinkBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")
		
		# set block's parameters
		# todo: add pranthesis to next strings if they are expressions
		self.amplitude_val = get_property(self.xml, "Amplitude")
		self.period_val = get_property(self.xml, "Period")
		self.pulsewidth_val = get_property(self.xml, "PulseWidth")
		self.phasedelay_val = get_property(self.xml, "PhaseDelay")
	
	def add_state_ports(self):
		pi = Port(self.get_id(), 0, "sin", 1)
		self.inport.append(pi)
		po = Port(self.get_id(), 0, "sout", 1)
		self.outport.append(po)
		
	def update_arities(self):
		self.outport[0].set_arity(1)
		return 0
	
	def split_block(self):
		return [SimulinkDiscretePulseGeneratorSingle(self, 0), SimulinkDiscretePulseGeneratorSingle(self, 1)]
	
	def set_init(self):
		for p in self.inport:
			if p.get_direction() == "sin":
				p.set_value(0.0)
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		func_type = "::(real \<times> real => bool) => real => bool"
		func = iBasic("PulseGeneratorS", [self.period_val, self.phasedelay_val, self.pulsewidth_val, self.amplitude_val] + self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
				
		return func

# Structure for PulseGenerator veriant with one output
class SimulinkDiscretePulseGeneratorSingle(SimulinkBlock):
	def __init__(self, pulsegen, select_output):
		SimulinkBlock.__init__(self)

		'''
		if select_output != 1:
			self.params = pulsegen.get_params_list()
		else:
			self.params = []
		'''
		
		self.params = pulsegen.get_params_list()
		self.pulsegen = pulsegen
		self.select_output = select_output
		self.outport = [pulsegen.outport[select_output]]
		self.inport = pulsegen.inport
		self.parent = pulsegen.parent
		self.name = pulsegen.name + ["A", "B"][select_output]
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
			
		if self.select_output == 0:
			func_type="::(real => bool) => real => bool"
			func = iBasic("PulseGeneratorSA", [self.pulsegen.period_val, self.pulsegen.phasedelay_val, self.pulsegen.pulsewidth_val, self.pulsegen.amplitude_val] + self.get_params_list())
		else:
			func_type="::(real => bool) => real => bool"
			func = iBasic("PulseGeneratorSB", [self.pulsegen.period_val, self.pulsegen.phasedelay_val, self.pulsegen.pulsewidth_val] + self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
						
		return func

# Structure for a Step block		
class SimulinkStep(SimulinkStateBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")

		# set block's parameters
		self.time_val   = get_property(self.xml, "Time")
		self.before_val = get_property(self.xml, "Before")
		self.after_val  = get_property(self.xml, "After")

	def split_block(self):
		return [SimulinkStepSingle(self, 0), SimulinkStepSingle(self, 1)]
		
	def update_arities(self):
		self.set_inport_arity(0, 1)
		self.set_outport_arity(0, 1)
		self.set_outport_arity(1, 1)
		return 0

	def set_init(self):
		self.set_port_init_val(0.0)
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		func_type = "::(real \<times> real => bool) => real => bool"
		func =  iBasic("StepT", [self.time_val, self.before_val, self.after_val] + self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
		
		return func

# Structure for Step variant with one output
class SimulinkStepSingle(SimulinkBlock):
	def __init__(self, step, select_output):
		SimulinkBlock.__init__(self)
		
		'''
		if select_output == 0:
			self.params = step.get_params_list()
		'''

		self.params = step.get_params_list()
		self.step = step
		self.select_output = select_output
		self.outport = [step.outport[select_output]]
		self.inport = step.inport
		self.parent = step.parent
		self.name = step.name + ["A", "B"][select_output]
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		if self.select_output == 0:
			func_type="::(real => bool) => real => bool"
			func = iBasic("StepTA", [self.step.time_val, self.step.before_val, self.step.after_val] + self.get_params_list())
		else:
			func_type="::(real => bool) => real => bool"
			func = iBasic("StepTB", self.get_params_list())

		if not opt_generic:
			return "("+str(func)+func_type+")"
						
		return func


# Structure for a SignalGenerator block
class SimulinkSignalGenerator(SimulinkStateBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")

		# set block's parameters
		self.ampl = 0 - float(get_property(self.xml, "Amplitude"))
		self.freq = get_property(self.xml, "Frequency")
		self.form  = get_property(self.xml, "WaveForm")
		self.unit = get_property(self.xml, "Units")
	
	def update_arities(self):
		self.set_inport_arity(0, 1)
		self.set_outport_arity(0, 1)
		self.set_outport_arity(1, 1)
		return 0

	def set_init(self):
		self.set_port_init_val(0.0)

	def split_block(self):
		return [SimulinkSignalGeneratorSingle(self, 0), SimulinkSignalGeneratorSingle(self, 1)]			

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())	
			
		st_var = self.inport[0].get_var_name()
		
		if self.unit == "Hertz":
			arg = "2*s_pi*"
		else:
			arg = ""
		if self.form == "square":
			fcn = "siggen_square" 
		
		expr1 = "%s * %s (%s%s*%s*%s)" %(self.ampl, fcn, arg, self.freq, st_var, self.get_params())
		expr2 = "%s+1" %(st_var)
		expr = "(%s, %s)" %(expr1, expr2)
		
		return iFunc([st_var], expr, opt_fun)
		

# Structure for SignalGenerator variant with one output		
class SimulinkSignalGeneratorSingle(SimulinkBlock):
	def __init__(self, sg, select_output):
		SimulinkBlock.__init__(self)
		if select_output == 0:
			self.params = sg.get_params_list()
			#self.params = step.get_params()
		self.sg = sg
		self.select_output = select_output
		self.outport = [sg.outport[select_output]]
		self.inport = sg.inport
		self.parent = sg.parent
		self.name = sg.name + ["A", "B"][select_output]
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			if self.select_output == 0:
				return iBasic(self.get_qualified_name(), self.get_params_list())
			else:
				return iBasic(self.get_qualified_name())

		st_var = self.inport[0].get_var_name()
		if self.select_output == 0:		
			if self.sg.unit == "Hertz":
				arg = "2*s_pi*"
			else:
				arg = ""
			if self.sg.form == "square":
				fcn = "siggen_square" 
		
			expr = "%s * %s (%s%s*%s*%s)" %(self.sg.ampl, fcn, arg, self.sg.freq, st_var, self.get_params())
			return iFunc([st_var], expr, opt_fun)
		else:
			expr = "%s+1" %(st_var)
			return iFunc([st_var], expr, opt_fun)	
		
		
		
		
		
		
		
		
		
		
		
		
		
