#!/usr/bin/env python


import subprocess
import sys
import os
import time
import fileinput
import re
import platform
from copy import deepcopy
from utilities.simulink_theories_path import *
from os.path import isfile, abspath, join, isabs, normpath, realpath

# TODO: make an options class

comp_options = [
	([("fp",      "feedback-parallel translation"),
	  ("ic",      "incremental translation"),
	  ("ici",     "incremental translation with intermediate outputs"),
	  ("nfb",     "feedbackless translation")], "nfb"),
	([("flat",    "flatten diagram for translation")], False),
	([("const",   "generates parameter types for constants (to be used in case of \"Additional types\" error given by Isabelle)")], False),
	([("type <Isabelle_type>", "instantiate all generic types with the parameter of this option. Parameter can be any valid Isabelle type, e.g., bool, int, real, int \<times> real, etc.")], False),
	([("generic", "use generic types for integrators, ..., Boolean blocks")],  False)]
	#([("def", "outputs the predicate transformers for unfolding")],  False),
		
cc_options = [
	("p", "path to isabelle batch file (usually Isabelle/bin)")]

	
def usage():
	comp_opt = " ".join(["[-%s]" % o for o in ["|".join([y[0] for y in x[0]]) for x in comp_options]])
	cc_opt = "-p <path/to/isabelle/bin>"
	print
	print "Usage: simcc.py %s <filename.slx> [<filename.thy>] %s" % (comp_opt, cc_opt)
	
	for o in comp_options:
		for op in o[0]:
			default = " (default)" if op[0] == o[1] else ""
			print "     -%s:\t%s%s" % (op[0], op[1], default)
	for o in cc_options:
		print "     -%s:\t%s" % (o[0], o[1])
		
	print
	exit(1)


def check_comp_options(argv):
	
	o = comp_options[0]
	labels = ["-" + x[0] for x in o[0]]
	sel = [x for x in argv if x in labels]
	if len(sel) > 1:
		print
		print "Error: At most one of the options %s must given" % (", ".join(labels))
		usage()
				

def rnpath(p): return realpath(normpath(p))

def extractFormulas(out):
	level = 0
	pos = []
	for i in range(len(out)):
		if out[i] == "(":
			level = level + 1
		if out[i] == ")":
			level = level - 1
		if out[i] == "," and level == 1:
			pos.append(i)
	formulas = []
	formulas.append(out[1:pos[0]])
	for i in range(1,len(pos)):
		formulas.append(out[pos[i-1]+1:pos[i]])
	formulas.append(out[pos[len(pos)-1]+1:len(out)-1])	
	return formulas

def extractVars(outputFile):
		out = open(outputFile, "r")
		results = []
		curKey = -1
		started = False
		for line in out:
			line = line.strip()
			#print line
 
			if not line: 
				continue
			elif line.startswith('simplify_RCRS'):
				started = True
				curKey = curKey + 1
				results.append(line)
			elif line.startswith('ML'):
				started = False
				continue
			elif started:
				results[curKey] = results[curKey] + line + "\n"
		out.close()
		f = results[len(results)-1]
		aux=f.split("\"")
		inpts_svars = aux[3]
		if inpts_svars.startswith("(") and inpts_svars.endswith(")"):
			inpts_svars=(inpts_svars.lstrip("(")).rstrip(")")
		outpts = aux[5]
		if outpts.startswith("(") and outpts.endswith(")"):
			outpts=(outpts.lstrip("(")).rstrip(")")
		
		inpts=[]
		svars=[]
		aux = [v.strip() for v in inpts_svars.split(",")]
		for v in aux:
			if v.startswith("si_"):
				svars.append(v)
			else:
				inpts.append(v)
		
		if len(inpts) == 1 and inpts[0].startswith("u_"):
			inpts=[]
		
		outpts = outpts.split(",")
		outpts_svars = [v.strip() for v in outpts]
		
		outpts = []
		for v in outpts_svars:
			if v.startswith("so_"):
				continue
			else:
				outpts.append(v)

		outpts_svars = ["si_"+v.replace("so_", "") + "'" if v.startswith("so_") else v for v in outpts_svars]

						
				
		return (inpts, outpts, svars, outpts_svars)



def extractAssert(out):
	s = re.match(r'{\.(.+)\.(.+)\.}', out, flags=re.DOTALL)
	if s:
		s = s.group(2)
	else:
		s = "True"
	return s.replace("?","")
	

def extractUpdate(out, var):
	
	ok = True

	#print out
	
	s = re.search(r'\[-\s+(.+?)\s+\\<leadsto>\s+(.+?)\s+-\]', out, flags=re.DOTALL) #
	#print s
	if s:
		s = s.group(2)
	else:
		ok = False

	if len(var) == 1 and not s:
		s = re.search(r'\[.([^.]+).\]', out) 
		if s:
			s = s.group(1)
			ok = True

	if not ok:
		print "Error: model might contain algebraic loops. Please check the diagram in Simulink."
		print
		exit (6)

	s=s.replace("\n"," ").strip()

	if len(var) > 1:
		val = extractFormulas(s)
		#val = s.split(',')
		if len(val) != len(var):
			print "Error: parsing error. Check theory file in GUI."
			print
			exit(7)
	else:
		val = [s]

	
	if val[0].count("(") != val[0].count(")"):
		val[0] = val[0][1:]

	if val[len(val)-1].count("(") != val[len(val)-1].count(")"):
		val[len(val)-1] = val[len(val)-1][:-1]

	r = []
	for i in range(len(val)):
		r.append(var[i] + " := " + val[i].strip().replace("?",""))
	res = " and\n\t ".join(r)
	return res

def extractSimp(out,outputfile,opt_args):
	
	#print out
	#print
	
	compStart = re.findall(r'val start_time_comp = (\d*.\d*): Time.time', out)
	compEnd = re.findall(r'val end_time_comp = (\d*.\d*): Time.time', out)
	printStart = re.findall(r'val start_time_print = (\d*.\d*): Time.time', out)
	printEnd = re.findall(r'val end_time_print = (\d*.\d*): Time.time', out)
 
	error = False
 
	if compStart and compEnd:
		dtComp = float(compEnd[0]) - float(compStart[0])
	else:
		error = True
 
	if printStart and printEnd:
		dtPrint = float(printEnd[0]) - float(printStart[0])
	else:
		error = True
 
	if error:
		# check for message "### Additional type variable(s) in specification" => use -type option
		#addType = re.search(r"### Additional type variable\(s\) in specification", out)
		addType = re.findall(r"Type unification failed: Clash of types .+", out) #\"[a-zA-Z_=>\s]+\"\s+and\s+\"[a-zA-Z_=>\s]+\"", out)
		types = None
		if addType:
			types = re.findall(r"\"([a-zA-Z_=>\s]+)\"", addType[0])
		if types:
			if "_ itself" in types:
				print 'Error: possible trapped variables cause an error at Isabelle execution. Use option -const.'
			else:
				#print addType
				#print 'Error: output variables set to reals while at least one is %s. Remove option -reals.' %([t for t in types if t != 'real'][0].strip())
				print 'Error: an error has occured at Isabelle execution. Use option -const and/or -generic.'
		#if addType:
		#	print 'Error: Possible trapped variables cause an error at Isabelle execution. Use option -type.'
		else:
			print 'Error: an error has occured at Isabelle execution. Check theory file in GUI.'
		return
		
	
 
	# extract model
 
	# extract simplification for the formula with reals

	# print out

	results = {}
	curKey = ''
	started = False
	ok = True
	auxString = ''
	for line in out.split('\n'):
		line = line.strip()
 
		if not line: 
			continue
		elif line == 'Expansion/simplification result':
			started = True
		elif line.startswith('val end_time_print ='):
			continue #break
		elif line.startswith('###'):
			continue
		elif line.startswith('val it'):
			break
		elif started:
			auxString = auxString + line
			m = re.match(r'([a-zA-Z][a-zA-Z0-9_\'()?\s.]+)\s*=\s*(.*)', auxString)
			if not m and ok:
				continue
			elif m and ok:
				m = [m.group(1), m.group(2)]
				#print m
				#print m
				curKey, v = m
				curKey = curKey.strip()
				#print curKey
				results[curKey] = v
				ok=False
			else:
				results[curKey] += line
				results[curKey] += "\n"
			'''
			m = re.findall(r'([a-zA-Z0-9_\'()?\s]+)\s*=\s*(.*)', line)
	
			if m:
				#print m
				curKey, v = m[0]
				curKey = curKey.strip()
				results[curKey] = v 
			else:
				results[curKey] += line
				results[curKey] += " "
				#results[curKey] += line.strip()
			'''
	#print results
	assert(1 <= len(results.items()) <= 2)
	#assert(len(results.items()) >= 1)
	
	res = None
	res_real = None
	
	for k, v in results.items():
		v = v.strip()
		if "_real" in k:
			res_real = v
		else:
			res = v
		#print k, '->', v
		#print

	'''
	aux = re.findall(r'Expansion/simplification result\s+[a-zA-Z0-9_\'()?\s]+\s+=\s+(.+)\s+val end_time_print = \d*.\d*: Time.time', out, flags=re.DOTALL)
	res = aux[0].strip()
	'''

	#print res
 	e = re.search(r'feedback', res)
	if e:
		print
		print 'Error: model might contain algebraic loops. Please check the diagram in Simulink.'
		print
		exit(3)


	print 'Done: %ss' %dtComp
	print

	addType = re.search(r"### Additional type variable\(s\) in specification", out)
	if addType and not("-const" in opt_args):
		print
		print 'Warning: trapped variables detected in the formula. For accurate results, use option -const.'
		print
	
	print "Results:"
	
		
	if res == "\<bottom>" or res_real == "\<bottom>":
		print '- Incompatibility found in the model. Check theory file in GUI. '
		return
	else:
		print '- All components are compatible.'

	v = extractVars(outputfile)
	a = extractAssert(res)
	u = extractUpdate(res,v[3])


	print "- Input variables: ", ('None' if len(v[0])==0 else ", ".join(v[0]))
	print "- Output variables: ", ('None' if len(v[1])==0 else ", ".join(v[1]))
	print "- State variables: ", ('None' if len(v[2])==0 else ", ".join(v[2]))
	if "?dt" in res:
		print "- Parameter: dt"
	if res_real != None:
		print "- Assert: ", extractAssert(res_real)	
		print "- Update: ", extractUpdate(res_real,v[3])	
	else:
		print "- Assert: ", extractAssert(res)	
		print "- Update: ", extractUpdate(res,v[3])	
	#print 'Simplified formula:\n%s' %res
	#print 
	#if res_real != None:
	#	print 'Simplified formula for reals:\n%s' %res_real
	#	print
	#print 'Formula printing time: %fs' %dtPrint
	#print
	
	return
 

def main():

	comp_opt = ['-fp', '-ic', '-ici', '-nfb', '-flat', '-const', '-type', '-generic']
	err_comp_opt = ['-iter','-sim','-equiv']
	
	# obtain the arguments from the command
	args = sys.argv[:]
	del args[0]
	
	# obtain path to isabelle_process  
	isabelle_path = None
	
	i = 0
	while i < len(args):
		x = args[i]
		if x == '-p':
			del args[i]
			try:
				isabelle_path = rnpath(args[i])
				del args[i]
			except:
				print "Error: The path for isabelle_process is incorrect\n"
				exit(2)
		else:
			i += 1

	syst = platform.system()

	if isabelle_path is None and syst != 'Windows':
		print
		print "Error: Missing path to isabelle_process"
		usage()

	# obtain slx file and check validity of arguments

	args1 = deepcopy(args)

	check_comp_options(args1)
	
	i = 0
	while i < len(args1):
		x = args1[i]
		if x in comp_opt: 
			comp_opt.remove(x)
			del args1[i]
			if x =="-type":
				del args1[i]
		elif x in err_comp_opt:
			print
			print "Warning: The option %s is not handled. Removed from compilation options." %x
			del args1[i]
			args.remove(x)
		else:
			i += 1

	#print args1
		
	if len(args1) > 2:
		print
		print "Error: Unknown options"
		usage()
		
	if len(args1) == 0:
		print
		print "Error: Missing Simulink file"
		usage()
		
	if len(args1) == 2:
		if (".slx" in args1[0] and not ".thy" in args1[1]) or (".thy" in args1[0] and not ".slx" in args1[1]):
			print
			print "Error: Incorrect options"
			usage()
			

	
	#print args1
	inputfile = ""
	if ".slx" in args1[0]:
		inputfile = args1[0]
	else:
		inputfile = args1[1]
	
	if len(args1) == 2:
		if ".thy" in args1[1]:
			outputfile = args1[1] 
		else:
			outputfile = args1[0] 
	else:
		outputfile = inputfile.replace(".slx",".thy")

	# obtain compiler and run it
	currentFolder =  os.path.dirname(os.path.realpath(__file__))
	
	compilerPath = rnpath(join(currentFolder, "simulink2isabelle.py"))

	print
	print "~~~Executing compiler~~~"
	print
	
	args2 = deepcopy(args)
	args2.remove(inputfile)
	if outputfile in args2: args2.remove(outputfile)
	args2 = [inputfile, outputfile] + args2
	
	print "Running compiler..."
	print
	
	time_start = time.time()
	
	retval = subprocess.call(['python', compilerPath] + args2)
	
	time_elapsed = (time.time() - time_start)
	
	if retval != 0:
		r = deepcopy(args)
		r.remove(inputfile)
		if outputfile in r:
			r.remove(outputfile)
		opt_args = " ".join(x for x in r)
		print  "Error: model %s with compiling options %s failed. Please check the model." %(inputfile, opt_args)
		print 
		exit(1)
	print
	print "Done: %ss" %time_elapsed
	print
	
	
	# build Simulink session 
	print
	print "~~~Creating Isabelle session~~~"
	print
	currentFolder =  os.path.dirname(rnpath(__file__))
	thFolder = rnpath(join(get_isa_theories_path(currentFolder), "Simulink"))
	if syst == 'Windows':
		i_path = "isabelle"
	else:
		i_path = rnpath(join(isabelle_path, "isabelle"))
	print "Building Simulink libraries"
	print "Please wait..."
	
	time_start = time.time()
		
	#subprocess.call([i_path, "build", "-D", thFolder, "-b"])
	#print i_path
	#print rnpath(thFolder)
	rout = subprocess.check_output([i_path, "build", "-D", rnpath(thFolder), "-b"])
	
	time_elapsed = (time.time() - time_start)

	print
	print "Done: %ss" %time_elapsed
	print

	
	# execute isabelle_process on the thy file 
	print
	print "~~~Executing Isabelle on theory~~~"
	print
	if syst == 'Windows':
		ip_path = "isabelle"
	else:
		ip_path = rnpath(join(isabelle_path, "isabelle"))
	command = 'use_thy \"%s\"; exit 0;' %outputfile.replace(".thy", "")
	session = 'Simulink'
	
	print "Performing expansion/simplification and checking compatibility of components"
	print
	
	#rout = subprocess.check_output([ip_path, "process", "-e", command])
	try:
		rout = subprocess.check_output([ip_path, "process", "-T", outputfile.replace(".thy", ""), "-l", "Simulink", "-d", rnpath(thFolder)])
		extractSimp(rout,outputfile,args2)
	except subprocess.CalledProcessError as e:
		print "The model %s could not be processed by Isabelle.\nCheck the theory for errors." %(outputfile)
		exit(1)
	# extract information about compatiblity and expanded/simplified formula

	
	print
	
	#print rout
	
	
main()
