# -*- coding: utf-8 -*-

from blocks.simulink_system import *
from utilities.simulink_utilities import *
import utilities.simulink_globals 

class SerialSingle(SimulinkBlock):
	def __init__(self, A, B):
		SimulinkBlock.__init__(self)
		# to treat Skip differently
		
		self.name = "SerialSingle"
		
		outA = A.outport[0]
		inport_B = [v for v in B.inport if v != outA and v not in A.inport]
		
		self.A = A
		self.B = B
		'''if A.get_params() == "":
			self.set_params(B.get_params_list())
		else:
			self.set_params(A.get_params_list())
		'''	
		'''if A.get_params() == "":
			self.set_params(B.get_params())
		else:
			self.set_params(A.get_params())
		'''
		
		self.set_params(A.get_params_list())
		self.set_params(B.get_params_list())
		
		self.inport = A.inport + inport_B
		self.outport = B.outport
		self.skip = A.is_skip() and B.is_skip()
		
		if A.is_skip():
			self.basic = B.is_basic()
		elif B.is_skip():
			self.basic = A.is_basic()
		else:
			self.basic = False
			
	def is_rename(self):
		return self.A.is_skip() or self.B.is_skip()
		
	def get_rename_base(self):
		if self.A.is_skip():
			return self.B.get_rename_base()
		elif self.B.is_skip():
			return self.A.get_rename_base()
		else:
			return self

	def get_dependent_blocks(self):
		if self.is_basic():
			return []
		dep = []
		for A in [self.A.get_rename_base(), self.B.get_rename_base()]:
			if (not A.is_basic()) and isinstance(A, SerialSingle):
				dep.append(A.get_name())
				
		return dep
		
	def get_type(self):
		return "SerialSingle"


	def get_bblocks(self, res):
		#print self.get_qualified_name()
		if self.A.get_type() == 'SerialSingle':
			self.A.get_bblocks(res)
		else:
			if not self.A.is_skip() and self.A not in res:
				res.append(self.A)
			else:
				pass
				#print self.A.get_qualified_name()
		if self.B.get_type() == 'SerialSingle':
			self.B.get_bblocks(res)
		else:
			if not self.B.is_skip() and self.B not in res:
				res.append(self.B)
			else:
				pass
				#print self.A.get_qualified_name()
				
	
	def get_sample_dt(self, vals):
		if self.A.get_type() == 'SerialSingle':
			self.A.get_sample_dt(vals)
		else:
			if not self.A.is_skip() and self.A.get_type() in ["Delay", "UnitDelay"]:
				vals[self.A] = self.A.sample_dt
				#vals[self.A.get_qualified_name()] = self.A.sample_dt
				#vals.append(self.A.sample_dt)
		if self.B.get_type() == 'SerialSingle':
			self.B.get_sample_dt(vals)
		else:
			if not self.B.is_skip() and self.B.get_type() in ["Delay", "UnitDelay"]:
				vals[self.B] = self.B.sample_dt
				#vals[self.B.get_qualified_name()] = self.B.sample_dt
				#vals.append(self.B.sample_dt)
							
	def isabelle(self, opt_fun, opt_generic, flag):
		
		#todo: remove is_skip from blocks. It is done when creating Isabelle syntax in iProd, iComp, ...
		if self.is_skip():
			return iSkip()
			
		if self.is_rename() or self.is_basic() or flag:
			outA = self.A.outport[0]
			inport_B = [v for v in self.B.inport if v != outA]
			if not inport_B == []:
				a = iSwitchSplit([self.inport], [self.A.inport, inport_B], opt_fun)
				b = iProd(self.A.isabelle(opt_fun, opt_generic, False), iSkip())
				c = iSwitch([self.A.outport, inport_B], [self.B.inport], opt_fun)
				d = self.B.isabelle(opt_fun, opt_generic, False)
				return iComp(a, b, c, d)
			else:
				a = iSwitchSplit([self.inport], [self.A.inport], opt_fun)
				b = self.A.isabelle(opt_fun, opt_generic, False)
				c = iSwitch([self.A.outport], [self.B.inport], opt_fun)
				d = self.B.isabelle(opt_fun, opt_generic, False)
				return iComp(a, b, c, d)
				
		else:
			return iBasic(self.get_name(), self.get_params_list())

class ProdSingle(SimulinkBlock):
	def __init__(self, system):
		SimulinkBlock.__init__(self)
		self.name = "NFB_Model"
		self.blocks = system.blocks
		self.display_blocks = system.display_blocks
		self.display_blocks.append(self)
		self.set_params(system.get_params_list())
		#self.set_params(system.get_params())
		self.outport = [b.outport[0] for b in self.blocks if len(b.outport) > 0]
		
		self.inport = []
		for b in self.blocks:
			ports = [port for port in b.inport if port not in self.inport]
			self.inport.extend(ports)
		self.formula = self
		
	def get_type(self):
		return "ProdSingle"
	
	#todo: modify similarly to SerialSingle.get_dependent_blocks
	def get_dependent_blocks(self):
		dep = []
		for b in self.blocks:
		#for b in self.display_blocks:
			if isinstance(b, SerialSingle) and (not b.is_basic()):
				for x in [b.get_rename_base()]:
					if isinstance(x, SerialSingle) and (not x.is_basic()):
						dep.append(x.get_name())
		return dep
		'''
		if self.is_basic():
			return []
		dep = []
		for A in self.display_blocks:
			if (not A.is_basic()) and isinstance(A, SerialSingle):
				dep.append(A.get_name())
				
		return dep		
		'''
	def isabelle(self, opt_fun, opt_generic, flag):
	
		res = ""
		if flag:
			a = iSwitchSplit([self.inport], [b.inport for b in self.blocks], opt_fun)
			bisabelle = [b.isabelle(opt_fun, opt_generic, False) for b in self.blocks]
			prod = iProd(*bisabelle)
			#print "TTTEEEST"
			c = iSwitch([b.outport for b in self.blocks], [self.outport], opt_fun)
			#print "My Skip: ", c, c.is_skip()
			return iComp(a, prod, c)
		else:
			return iBasic(self.get_name(), self.get_params_list())
		
	def get_fblocks(self,res):	
		for b in self.display_blocks:
			if isinstance(b, SimulinkFunction):
				res.append(b)	
				
	def get_bblocks(self,res):	
		for b in self.blocks:
			if not isinstance(b, SerialSingle):
				if not isinstance(b, ProdSingle):
					if b not in res:
						res.append(b)	
			else:
				b.get_bblocks(res)
	
	def get_sample_dt(self, vals):
		for b in self.blocks:
			if not isinstance(b, SerialSingle):
				if not isinstance(b, ProdSingle):
					if b.get_type() in ["Delay", "UnitDelay"]:
						vals[b] = b.sample_dt
						#vals[b.get_qualified_name()] = b.sample_dt
						#vals.append(b.sample_dt)	
			else:
				b.get_sample_dt(vals)
		'''
		
		for b in self.display_blocks:
			if b.get_type() in ["Delay", "UnitDelay"]:
				vals.append(b.sample_dt)
				
		'''
		
class NoFeedbackComposition(SimulinkBlock):
	def __init__(self, system):
		SimulinkBlock.__init__(self)
		self.sid = system.sid
		self.name = ""
		self.params = []
		self.display_blocks = []
		
		new_blocks = []

		subsystem_index = 0
		for b in system.blocks:
			if isinstance(b, SimulinkSystem):
				nfb_block = NoFeedbackComposition(b)
				new_blocks.extend(nfb_block.blocks)
				if nfb_block.display_blocks != []:
					subsystem_index = subsystem_index + 1
					subsystem_prefix = "t%d" % subsystem_index
				for b in nfb_block.display_blocks:
					b.name = subsystem_prefix + b.name
				self.display_blocks.extend(nfb_block.display_blocks)
			else:
				new_blocks.append(b)
				if b.get_type() == 'Fcn':
					self.display_blocks.append(b)

		for b in new_blocks:
			if b.get_params() != "":
				self.set_params(b.get_params_list())
				#self.set_params("dt")
				#break
		
		self.del_inport = []
		for b in new_blocks:
			self.del_inport.extend(b.del_inport)
			if len(b.outport) > 1:
				#print b.name
				#print b.outport
				raise Exception('\nError:\n  The feedbackless method requires all blocks to be decomposable with 1-arity outports.\n  The block '+ b.name + ' is not currently handled.\n  Please try another translation strategy.\n') #all blocks must have at most one output port')
		
		current_blocks = new_blocks

		#remove blocks that are obsolete
		#print "delete ports:", self.del_inport
		
		blocks = [b for b in current_blocks if not b.outport or (b.outport[0] not in self.del_inport)]
		del_blocks = [b for b in current_blocks if b.outport and b.outport[0] in self.del_inport]
		current_blocks = blocks
		while del_blocks != []:
			in_del_ports = []
			for b in del_blocks:
				in_del_ports.extend(b.inport)
			
			in_ports = []
			for b in blocks:
				in_ports.extend(b.inport)
				
			del_ports = [b for b in in_del_ports if b not in in_ports]
			
			blocks = [b for b in current_blocks if not b.outport or b.outport[0] not in del_ports]
			del_blocks = [b for b in current_blocks if b.outport and b.outport[0] in del_ports]
			current_blocks = blocks
		#print "delete blocks:", del_blocks
		done = False
		tindex = 0
		while not done:
			done = True
			comp = False
			for b in current_blocks:
				if len(b.outport) > 0:
					outp = b.outport[0]
					if outp in b.inport:
						raise Exception('\nError:\n  An algebraic loop has been detected in the model.\n  Please check the model or try a different translation strategy.\n')
					update_blocks = [bl for bl in current_blocks if outp in bl.inport]
					if update_blocks != []:
						comp = True
						current_blocks = [bl for bl in current_blocks if (not outp in bl.inport) and bl != b]
						break
			if comp:
				done = False
				for x in update_blocks:
					#print "AAA:", b.get_id()
					block = SerialSingle(b, x)
					#if b.get_id() == "64":
						#print "64:", x.isabelle(True, True)
						#print "64:", block.isabelle(True, True)
					if(not block.is_basic() and not block.is_rename()):
						tindex = tindex + 1
						block.name = 't%d' % tindex
						self.display_blocks.append(block)

					current_blocks.append(block)
					
		
		self.blocks = current_blocks
		
	def get_id(self):
		return self.sid

