# -*- coding: utf-8 -*-

import pyparsing as PP
import xml.etree.cElementTree as ET
import utilities.simulink_globals


global ops
ops = set(['+', '*', '-', '/', '^', '=='])
global prec
prec = {'+':10, '*':20, '-':10, '/':20, '^':30, '==':5}

global fnArity
fnArity = {'sqrt' : 1, 'max' : 2, 'min' : 2, 'log' : 1, 'abs' :  1, 'sin' : 1, 'cos' : 1, 'exp': 1}

	
for fn in fnArity:
	ops.add(fn)
	prec[fn] = 9999 # over nine thousand

# method that ...
def inf2post(tokens):
	output = []
	stack = []
	for item in tokens:
		#pop elements while elements have lower precedence
		if item in ops:
			while stack and stack[-1] in ops and prec[stack[-1]] >= prec[item]:
				output.append(stack.pop())
			stack.append(item)
		#delay precedence. append to stack
		elif item == "(":
			stack.append("(")
		#flush output until "(" is reached
		elif item == ")":
			while stack and stack[-1] != "(":
				output.append(stack.pop())
			#should be "("
			#print 
			stack.pop()
		#operand. append to output stream
		else:
			output.append(item)
	#flush stack to output
	while stack:
		output.append(stack.pop())
	return output
	
	
# method that ...    
def post2tree(postfixList):
	tree = []
	for index, item in enumerate(postfixList):
		if item in ops:
			op2 = tree.pop()
			op1 = tree.pop()
			tree.append([item, op1, op2])
		else:
			tree.append(item)
	
	return tree[0]  
	

def xmlTraverse(tree, padding = ''):
	print padding, tree.tag, tree.attrib
	for child in tree:
		xmlTraverse(child, padding + '-')


def paIdentifier(s, l, t):
	pass


def paTerm(s, l, t):
	pass


def unaryMinusAction(s, l, t):
	t.insert(0, '0')


PP_identifier = PP.Word(PP.alphas + '_', PP.alphanums + '_').setParseAction(paIdentifier)
PP_spaces = PP.Word(' ')
PP_integer = PP.Combine( PP.Optional(PP.oneOf('+ -')) + PP.Optional(PP_spaces) + PP.Word(PP.nums) )
PP_real = PP.Combine( PP_integer + 
				   PP.Optional(PP.Literal('.') + PP.Optional(PP.Word(PP.nums)) ) +
				   PP.Optional( PP.CaselessLiteral('e') + PP_integer) ) 
				   
#PP_real.setParseAction(lambda s, l, t: ['mynumber(' + ''.join(t) + ')']) # TODO: should all reals be turned into 'mynumber' ?

PP_comment = PP.Word('%', PP.printables + ' \t')
#PP_comment.setParseAction(lambda s, l, t: '#' + t[0][1:])

PP_assignment = (PP_identifier + '=' + PP_real)
#PP_assignment.setParseAction(lambda s, l, t: '%s = %s' % (t[0], t[2]))
PP_assignment_string = (PP_identifier + '=' + PP_identifier)

PP_annotation_line = (PP_assignment | PP_comment | PP_assignment_string)

PP_expr = PP.Forward()

# matlab guy uses () but slx file has [] - actually, another slx file has ()...
# context is necessary to distinguish between fn application and array indexing in case of ()
# so, just parse array indexing as fn application, but treat them differently during semantic actions
# ^ NO, this causes problems in wrapping literals with 'mynumber'. use a priori known names for functions instead ;)

#PP_fn_name = PP.Literal('sqrt') | PP.Literal('pow') | PP.Literal('f')

PP_fn_name = PP.Forward()


for fn in fnArity:
    PP_fn_name |= PP.Literal(fn)


PP_arr_indexing = PP_identifier + PP.oneOf('[ (') + PP.Word(PP.nums) + PP.oneOf('] )') 
PP_arr_indexing.setParseAction(lambda s, l, t: [t[0] + str(int(t[2]))]) # remember: python lists are 0-indexed ;) --> modified for isabelle

PP_fn_application = PP_fn_name + '(' + PP.Optional( PP_expr + PP.ZeroOrMore(PP.Literal(',') + PP_expr) ) + ')'
#PP_fn_application.setParseAction(lambda s, l, t: ['math.' + ''.join(t)])
PP_fn_application.setParseAction(lambda s, l, t: (t[:] + ['dummy_2nd_arg']) if fnArity[t[0]] == 1 else [t[0]] + list(IT.takewhile(lambda c: c != ',', t[1:])) + [')'] + ['(']+ list(IT.dropwhile(lambda c: c != ',', t[1:]))[1:])

PP_atom = PP.Forward()
PP_atom << (PP_real | PP_fn_application | PP_arr_indexing | PP_identifier | PP.Literal('(') + PP_expr + PP.Literal(')') | (PP.Literal('-') + PP_atom).setParseAction(unaryMinusAction))  # DO NOT SUPPRESS PARENTHESES, WE NEED THEM!!!


#PP_atom = PP_real | PP_fn_application | PP_arr_indexing | PP_identifier | PP.Literal('(') + PP_expr + PP.Literal(')') # DO NOT SUPPRESS PARENTHESES, WE NEED THEM!!!

PP_factor = PP.Forward() # this way, ^ associativity is right-to-left
#PP_factor << PP_atom + PP.ZeroOrMore(PP.Literal('^').setParseAction(lambda s, l, t: ['**']) + PP_factor) 


PP_factor << PP_atom + PP.ZeroOrMore(PP.Literal('^') + PP_factor) 


PP_term = PP_factor + PP.ZeroOrMore(PP.oneOf('* /') + PP_factor)
PP_term.setParseAction(paTerm)


PP_cmp = PP_term + PP.ZeroOrMore(PP.oneOf('+ -') + PP_term)

PP_expr << PP_cmp + PP.ZeroOrMore(PP.Literal('==') + PP_cmp)

#PP_expr.setParseAction(lambda s, l, t: ' '.join(t))

#print dict(PP_expr)
#print dir(PP_expr)
#string = '1.0 - u(1)/(u(2) + (u(2)==0)*eps)'
#print post2tree(inf2post(PP_expr.parseString(string)))

# needed ?
def make_property_dict(block, property_name):
	
	bprop = {}
	
	for p in block.findall('./P'):
		bprop[p.attrib['Name']] = p.text
	
	if 'BlockType' in elem.attrib:     
		for k, v in block_parameter_defaults[block.attrib['BlockType']].iteritems():
			if k not in bprop:
				bprop[k] = v
	
	return bprop


# method that ...
def get_property(elem, property_name):
	
	for p in elem.findall('./P'):
		if p.attrib['Name'] == property_name:
			return p.text

	if 'BlockType' in elem.attrib:        
		for k, v in utilities.simulink_globals.block_parameter_defaults[elem.attrib['BlockType']].iteritems():
			if k == property_name:
				return v
	
	return None


def get_instance_data_property(elem, property_name):
	
	for p in elem.findall('./InstanceData/P'):
		if p.attrib['Name'] == property_name:
			return p.text

	return None

# method that ...
def set_property(elem, property_name, new_value):
	
	for p in elem.findall('./P'):
		if p.attrib['Name'] == property_name:
			p.text = new_value
			break
	else:
		pass #create new property here
		new_p_node = ET.fromstring('<P Name="%s">%s</P>' % (property_name, new_value))
		elem.append(new_p_node)


# needed ?
def line_signal(line):
	direct_parent = child2parent[line]
	if direct_parent.tag in ['Line', 'Branch']:
		return line_signal(direct_parent)
	src = line.find("./P[@Name='Src']").text
	
	''' this is not necessary after all...
	# TODO: check if source SID is a subsystem and do stuff... - DONE!
	sid = src.split('#')[0]
	port = src.split(':')[1]
	src_elem = sid2element[sid]
	if src_elem.attrib['BlockType'] == 'SubSystem':
		outs = src_elem.findall("./System/Block[@BlockType='Outport']")
		for o in outs:
			if get_property(o, 'Port') == port:
				src = "{}#out:{}".format(o.attrib['SID'], port)
				break
		else:
			print src_elem.attrib['Name']
			for o in outs:
				print get_property(o, 'Port')
			print port
			raise Exception("subsystem doesn't have outports?...")
	#'''
	
	# TODO: what about other types of ports? do they affect this? - NO
	return src


# needed for ?        
def get_in_out_sigs_helper(block, src_dst, in_out):
	 
	ret = {}
	bsid = block.attrib['SID']
	
	for line in all_lines:
		# if line has Dst/Src property and if Dst/Src starts with SID...
		dst = get_property(line, src_dst)
		if dst != None and dst.split('#')[0] == bsid:
			sig_id = dst.split('#')[1].split(':')[0]
			if sig_id not in ret:
				ret[sig_id] = []
			ret[sig_id].append(line)
	
	if in_out in ret:
		ret[in_out].sort(key = lambda l: int(get_property(l, src_dst).split(':')[1]))
	
	for k in ret:
		ret[k] = map(line_signal, ret[k])
	
	return ret


# method that ...	
def get_incoming_signals(block):
	
	ret = get_in_out_sigs_helper(block, 'Dst', 'in')
	
	#TODO: signals propagated without lines (e.g. in/out-ports, gotos, froms etc...) - DONE!
	btype = block.attrib['BlockType']
	
	if btype == 'Inport':
		parent = child2parent[child2parent[block]]
		if parent.tag == 'Block' and parent.attrib['BlockType'] == 'SubSystem':
			port = get_property(block, 'Port')
			ret['in'] = [get_incoming_signals(parent)['in'][int(port) - 1]]
	elif btype == 'From':
		ret['in'] = [get_property(block, 'GotoTag')]
	
	return ret


# method that ...
def get_outgoing_signals(block):
	
	ret = get_in_out_sigs_helper(block, 'Src', 'out')
	
	#TODO: signals propagated without lines (e.g. in/out-ports, gotos, froms etc...) - DONE!
	btype = block.attrib['BlockType']
	
	if btype == 'Outport':
		parent = child2parent[child2parent[block]]
		if parent.tag == 'Block' and parent.attrib['BlockType'] == 'SubSystem':
			port = get_property(block, 'Port')
			ret['out'] = [get_outgoing_signals(parent)['out'][int(port) - 1]]
	elif btype == 'Goto':
		ret['out'] = [get_property(block, 'GotoTag')]
	
	return ret
